from datetime import timedelta

import pandas as pd
from couchbase.auth import PasswordAuthenticator
from couchbase.cluster import Cluster
from couchbase.exceptions import UnAmbiguousTimeoutException
from couchbase.options import ClusterOptions
from couchbase.exceptions import KeyspaceNotFoundException, CouchbaseException

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities import log
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)


logger = log.getLogger(__name__)


class CouchbaseHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Couchbase statements.
    """

    name = "couchbase"
    # TODO: Check the timeout value with the sdk default time
    DEFAULT_TIMEOUT_SECONDS = 60

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.connection_data = kwargs.get("connection_data")

        self.scope = self.connection_data.get("scope") or "_default"

        self.bucket_name = self.connection_data.get("bucket")
        self.cluster = None

        self.is_connected = False

    def connect(self):
        """
        Set up connections required by the handler.

        Returns:
            The connected cluster.
        """
        if self.is_connected:
            return self.cluster

        auth = PasswordAuthenticator(
            self.connection_data.get("user"),
            self.connection_data.get("password"),
            # NOTE: If using SSL/TLS, add the certificate path.
            # We strongly reccomend this for production use.
            # cert_path=cert_path
        )

        options = ClusterOptions(auth)

        if 'cloud.couchbase.com' in self.connection_data.get("host"):
            options.apply_profile('wan_development')

            endpoint = f"couchbases://{self.connection_data.get('host')}"
        else:
            endpoint = f"couchbase://{self.connection_data.get('host')}"

        cluster = Cluster(
            endpoint,
            options,
        )

        try:
            cluster.wait_until_ready(timedelta(seconds=self.DEFAULT_TIMEOUT_SECONDS))
            self.is_connected = cluster.connected
            self.cluster = cluster
        except UnAmbiguousTimeoutException:
            self.is_connected = False
            raise

        return self.cluster

    def disconnect(self):
        """Close any existing connections
        Should switch self.is_connected.
        """
        if self.is_connected is False:
            return
        self.is_connected = self.cluster.connected
        return

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the Couchbase bucket
        :return: success status and error message if error occurs
        """
        result = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            cluster = self.connect()
            result.success = cluster.connected
        except UnAmbiguousTimeoutException as e:
            logger.error(
                f'Error connecting to Couchbase {self.connection_data["bucket"]}, {e}!'
            )
            result.error_message = str(e)

        if result.success is True and need_to_close:
            self.disconnect()
        if result.success is False and self.is_connected is True:
            self.is_connected = False
        return result

    def native_query(self, query: str) -> Response:
        """Execute a raw query against Couchbase.

        Args:
            query (str): Raw Couchbase query.

        Returns:
            HandlerResponse containing query results.
        """
        self.connect()
        bucket = self.cluster.bucket(self.bucket_name)
        cb = bucket.scope(self.scope)

        data = {}
        try:
            for collection in cb.query(query):
                for collection_name, row in collection.items():
                    if isinstance(row, dict):
                        for k, v in row.items():
                            data.setdefault(k, []).append(v)
                    else:
                        for k, v in collection.items():
                            data.setdefault(k, []).append(v)

            response = Response(
                RESPONSE_TYPE.TABLE, pd.DataFrame(data) if data else RESPONSE_TYPE.OK
            )
        except CouchbaseException as e:
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e.error_context.first_error_message),
            )

        return response

    def query(self, query: ASTNode) -> Response:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """
        return self.native_query(query.to_string())

    def get_tables(self) -> Response:
        """
        Get a list with of collection in database
        """
        cluster = self.connect()
        bucket = cluster.bucket(self.bucket_name)
        unique_collections = set()
        for scope in bucket.collections().get_all_scopes():
            for collection in scope.collections:
                unique_collections.add(collection.name)
        collections = list(unique_collections)
        df = pd.DataFrame(collections, columns=["TABLE_NAME"])
        response = Response(RESPONSE_TYPE.TABLE, df)

        return response

    def get_columns(self, table_name) -> Response:
        """Returns a list of entity columns
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse: shoud have same columns as information_schema.columns
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-columns-table.html)
                Column 'COLUMN_NAME' is mandatory, other is optional. Hightly
                recomended to define also 'DATA_TYPE': it should be one of
                python data types (by default it str).
        """

        response = Response(False)

        cluster = self.connect()
        bucket = cluster.bucket(self.bucket_name)
        cb = bucket.scope(self.scope)

        try:
            q = f"SELECT * FROM `{table_name}` limit 1"
            row_iter = cb.query(q)
            data = []
            for row in row_iter:
                for k, v in row[table_name].items():
                    data.append([k, type(v).__name__])
            df = pd.DataFrame(data, columns=["Field", "Type"])
            response = Response(RESPONSE_TYPE.TABLE, df)
        except KeyspaceNotFoundException as e:
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=f"Error: {e.error_context.first_error_message}",
            )

        return response



from mindsdb.integrations.libs.ml_handler_process.create_validation_process import create_validation_process
from mindsdb.integrations.libs.ml_handler_process.create_engine_process import create_engine_process
from mindsdb.integrations.libs.ml_handler_process.update_engine_process import update_engine_process
from mindsdb.integrations.libs.ml_handler_process.describe_process import describe_process
from mindsdb.integrations.libs.ml_handler_process.predict_process import predict_process
from mindsdb.integrations.libs.ml_handler_process.update_process import update_process
from mindsdb.integrations.libs.ml_handler_process.learn_process import learn_process



from pydantic import BaseModel, Extra

DEFAULT_EMBEDDING_MODEL = "all-MiniLM-L6-v2"


class Parameters(BaseModel):
    embeddings_model_name: str = DEFAULT_EMBEDDING_MODEL
    text_columns: list = None
    use_gpu: bool = False

    class Config:
        extra = Extra.forbid



from typing import Dict, List

from langchain_core.tools import BaseTool
from sqlalchemy.orm.attributes import flag_modified
import pandas as pd

from mindsdb.interfaces.model.functions import PredictorRecordNotFound
from mindsdb.interfaces.storage.db import Predictor
from mindsdb.interfaces.model.model_controller import ModelController
from mindsdb.interfaces.skills.skills_controller import SkillsController
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.config import Config


class AgentsController:
    '''Handles CRUD operations at the database level for Agents'''

    def __init__(
        self,
        datahub,
        project_controller: ProjectController = None,
        skills_controller: SkillsController = None,
        model_controller: ModelController = None
    ):
        if project_controller is None:
            project_controller = ProjectController()
        if skills_controller is None:
            skills_controller = SkillsController()
        if model_controller is None:
            model_controller = ModelController()
        self.project_controller = project_controller
        self.skills_controller = skills_controller
        self.model_controller = model_controller
        self.datahub = datahub

    def get_agent(self, agent_name: str, project_name: str = 'mindsdb') -> db.Agents:
        '''
        Gets an agent by name.

        Parameters:
            agent_name (str): The name of the agent
            project_name (str): The name of the containing project - must exist

        Returns:
            agent (db.Agents): The database agent object
        '''

        project = self.project_controller.get(name=project_name)
        agent = db.Agents.query.filter(
            db.Agents.name == agent_name,
            db.Agents.project_id == project.id,
            db.Agents.company_id == ctx.company_id
        ).first()
        return agent

    def get_agent_by_id(self, id: int, project_name: str = 'mindsdb') -> db.Agents:
        '''
        Gets an agent by id.

        Parameters:
            id (int): The id of the agent
            project_name (str): The name of the containing project - must exist

        Returns:
            agent (db.Agents): The database agent object
        '''

        project = self.project_controller.get(name=project_name)
        agent = db.Agents.query.filter(
            db.Agents.id == id,
            db.Agents.project_id == project.id,
            db.Agents.company_id == ctx.company_id
        ).first()
        return agent

    def get_agents(self, project_name: str) -> List[dict]:
        """
        Gets all agents in a project.

        Parameters:
            project_name (str): The name of the containing project - must exist

        Returns:
            all-agents (List[db.Agents]): List of database agent object
        """

        all_agents = db.Agents.query.filter(
            db.Agents.company_id == ctx.company_id
        )

        if project_name is not None:
            project = self.project_controller.get(name=project_name)

            all_agents = all_agents.filter(db.Agents.project_id == project.id)

        return all_agents.all()

    def add_agent(
            self,
            name: str,
            project_name: str,
            model_name: str,
            skills: List[str],
            params: Dict[str, str] = {}) -> db.Agents:
        '''
        Adds an agent to the database.

        Parameters:
            name (str): The name of the new agent
            project_name (str): The containing project
            model_name (str): The name of the existing ML model the agent will use
            skills (List[str]): List of existing skill names to add to the new agent
            params (Dict[str, str]): Parameters to use when running the agent

        Returns:
            agent (db.Agents): The created agent

        Raises:
            ValueError: Agent with given name already exists, or skill/model with given name does not exist.
            NotImplementedError: Free users try to create an agent.
        '''

        config = Config()

        is_cloud = config.get('cloud', False)
        if is_cloud and ctx.user_class == 0:
            raise NotImplementedError('Free users cannot create agents. Please subscribe to MindsDB Pro.')

        if project_name is None:
            project_name = 'mindsdb'
        project = self.project_controller.get(name=project_name)

        agent = self.get_agent(name, project_name)

        if agent is not None:
            raise ValueError(f'Agent with name already exists: {name}')

        # Check if model exists.
        model_name_no_version, model_version = Predictor.get_name_and_version(model_name)
        try:
            self.model_controller.get_model(model_name_no_version, version=model_version, project_name=project_name)
        except PredictorRecordNotFound:
            raise ValueError(f'Model with name does not exist: {model_name}')

        agent = db.Agents(
            name=name,
            project_id=project.id,
            company_id=ctx.company_id,
            user_class=ctx.user_class,
            model_name=model_name,
            params=params,
        )
        skills_to_add = []
        # Check if given skills exist.
        for skill in skills:
            existing_skill = self.skills_controller.get_skill(skill, project_name)
            if existing_skill is None:
                raise ValueError(f'Skill with name does not exist: {skill}')
            skills_to_add.append(existing_skill)
        agent.skills = skills_to_add

        db.session.add(agent)
        db.session.commit()

        return agent

    def update_agent(
            self,
            agent_name: str,
            project_name: str = 'mindsdb',
            name: str = None,
            model_name: str = None,
            skills_to_add: List[str] = None,
            skills_to_remove: List[str] = None,
            params: Dict[str, str] = None):
        '''
        Updates an agent in the database.

        Parameters:
            agent_name (str): The name of the new agent, or existing agent to update
            project_name (str): The containing project
            name (str): The updated name of the agent
            model_name (str): The name of the existing ML model the agent will use
            skills_to_add (List[str]): List of skill names to add to the agent
            skills_to_remove (List[str]): List of skill names to remove from the agent
            params: (Dict[str, str]): Parameters to use when running the agent

        Returns:
            agent (db.Agents): The created or updated agent

        Raises:
            ValueError: Agent with name not found, agent with new name already exists, or model/skill does not exist.
        '''

        existing_agent = self.get_agent(agent_name, project_name=project_name)
        if existing_agent is None:
            raise ValueError(f'Agent with name not found: {agent_name}')

        if name is not None and name != agent_name:
            # Check to see if updated name already exists
            agent_with_new_name = self.get_agent(name, project_name=project_name)
            if agent_with_new_name is not None:
                raise ValueError(f'Agent with updated name already exists: {name}')

        # Check if model exists.
        if model_name is not None:
            model_name_no_version, model_version = Predictor.get_name_and_version(model_name)
            try:
                _ = self.model_controller.get_model(model_name_no_version, version=model_version, project_name=project_name)
                existing_agent.model_name = model_name
            except PredictorRecordNotFound:
                return ValueError(f'Model with name does not exist: {model_name}')

        # Check if given skills exist.
        new_skills = []
        for skill in skills_to_add:
            existing_skill = self.skills_controller.get_skill(skill, project_name)
            if existing_skill is None:
                raise ValueError(f'Skill with name does not exist: {skill}')
            new_skills.append(existing_skill)
        existing_agent.skills += new_skills

        removed_skills = []
        for skill in existing_agent.skills:
            if skill.name in skills_to_remove:
                removed_skills.append(skill)
        for skill_to_remove in removed_skills:
            existing_agent.skills.remove(skill_to_remove)

        if params is not None:
            # Merge params on update
            existing_params = existing_agent.params or {}
            existing_params.update(params)
            # Remove None values entirely.
            params = {k: v for k, v in existing_params.items() if v is not None}
            existing_agent.params = params
            # Some versions of SQL Alchemy won't handle JSON updates correctly without this.
            # See: https://docs.sqlalchemy.org/en/20/orm/session_api.html#sqlalchemy.orm.attributes.flag_modified
            flag_modified(existing_agent, 'params')
        db.session.commit()

        return existing_agent

    def delete_agent(self, agent_name: str, project_name: str = 'mindsdb'):
        '''
        Deletes an agent by name.

        Parameters:
            agent_name (str): The name of the agent to delete
            project_name (str): The name of the containing project

        Raises:
            ValueError: Agent does not exist.
        '''

        agent = self.get_agent(agent_name, project_name)
        if agent is None:
            raise ValueError(f'Agent with name does not exist: {agent_name}')
        db.session.delete(agent)
        db.session.commit()

    def get_completion(
            self,
            agent: db.Agents,
            messages: List[Dict[str, str]],
            project_name: str = 'mindsdb',
            tools: List[BaseTool] = None) -> pd.DataFrame:
        '''
        Queries an agent to get a completion.

        Parameters:
            agent (db.Agents): Existing agent to get completion from
            messages (List[Dict[str, str]]): Chat history to send to the agent
            project_name (str): Project the agent belongs to (default mindsdb)
            tools (List[BaseTool]): Tools to use while getting the completion

        Returns:
            pd.DataFrame (pd.DataFrame): Completion as a DataFrame

        Raises:
            ValueError: Agent's model does not exist.
        '''
        # Model needs to exist.
        model_name_no_version, version = db.Predictor.get_name_and_version(agent.model_name)
        try:
            _ = self.model_controller.get_model(model_name_no_version, version=version, project_name=project_name)
        except PredictorRecordNotFound:
            return ValueError(f'Model with name {agent.model_name} not found')

        if tools is None:
            tools = []
        predict_params = {
            # Underlying handler (e.g. Langchain) will handle default tools like mdb_read, mdb_write, etc.
            'tools': tools,
            'skills': [s for s in agent.skills],
            **(agent.params or {})
        }
        project_datanode = self.datahub.get(project_name)
        return project_datanode.predict(
            model_name=agent.model_name,
            data=messages,
            params=predict_params
        )



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __version__ as version, __description__ as description

try:
    from .dummy_data_handler import DummyHandler as Handler

    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = ''
name = 'dummy_data'
type = HANDLER_TYPE.DATA
icon_path = 'icon.svg'
permanent = False

__all__ = ['Handler', 'version', 'name', 'type', 'title', 'description', 'import_error', 'icon_path']



import datetime
import mimetypes
import os
import secrets
import threading
import traceback
import webbrowser
from distutils.version import LooseVersion
from pathlib import Path

import requests
from flask import Flask, url_for, make_response, request, send_from_directory
from flask.json import dumps
from flask_compress import Compress
from flask_restx import Api
from werkzeug.exceptions import HTTPException

from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.api.http.gui import update_static
from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.agents import ns_conf as agents_ns
from mindsdb.api.http.namespaces.analysis import ns_conf as analysis_ns
from mindsdb.api.http.namespaces.auth import ns_conf as auth_ns
from mindsdb.api.http.namespaces.chatbots import ns_conf as chatbots_ns
from mindsdb.api.http.namespaces.jobs import ns_conf as jobs_ns
from mindsdb.api.http.namespaces.config import ns_conf as conf_ns
from mindsdb.api.http.namespaces.databases import ns_conf as databases_ns
from mindsdb.api.http.namespaces.default import ns_conf as default_ns, check_auth
from mindsdb.api.http.namespaces.file import ns_conf as file_ns
from mindsdb.api.http.namespaces.handlers import ns_conf as handlers_ns
from mindsdb.api.http.namespaces.knowledge_bases import ns_conf as knowledge_bases_ns
from mindsdb.api.http.namespaces.models import ns_conf as models_ns
from mindsdb.api.http.namespaces.projects import ns_conf as projects_ns
from mindsdb.api.http.namespaces.skills import ns_conf as skills_ns
from mindsdb.api.http.namespaces.sql import ns_conf as sql_ns
from mindsdb.api.http.namespaces.tab import ns_conf as tab_ns
from mindsdb.api.http.namespaces.tree import ns_conf as tree_ns
from mindsdb.api.http.namespaces.views import ns_conf as views_ns
from mindsdb.api.http.namespaces.util import ns_conf as utils_ns
from mindsdb.interfaces.database.integrations import integration_controller
from mindsdb.interfaces.database.database import DatabaseController
from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.interfaces.storage import db
from mindsdb.metrics.server import init_metrics
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.json_encoder import CustomJSONEncoder
from mindsdb.utilities.ps import is_pid_listen_port, wait_func_is_true
from mindsdb.utilities.telemetry import inject_telemetry_to_static

logger = log.getLogger(__name__)


class Swagger_Api(Api):
    """
    This is a modification of the base Flask Restplus Api class due to the issue described here
    https://github.com/noirbizarre/flask-restplus/issues/223
    """
    @property
    def specs_url(self):
        return url_for(self.endpoint("specs"), _external=False)


def custom_output_json(data, code, headers=None):
    resp = make_response(dumps(data), code)
    resp.headers.extend(headers or {})
    return resp


def get_last_compatible_gui_version() -> LooseVersion:
    logger.debug("Getting last compatible frontend..")
    try:
        res = requests.get('https://mindsdb-web-builds.s3.amazonaws.com/compatible-config.json', timeout=5)
    except (ConnectionError, requests.exceptions.ConnectionError) as e:
        logger.error(f"Is no connection. {e}")
        return False
    except Exception as e:
        logger.error(f"Is something wrong with getting compatible-config.json: {e}")
        return False

    if res.status_code != 200:
        logger.error(
            f"Cant get compatible-config.json: returned status code = {res.status_code}"
        )
        return False

    try:
        versions = res.json()
    except Exception as e:
        logger.error(f"Cant decode compatible-config.json: {e}")
        return False

    current_mindsdb_lv = LooseVersion(mindsdb_version)

    try:
        gui_versions = {}
        max_mindsdb_lv = None
        max_gui_lv = None
        for el in versions['mindsdb']:
            if el['mindsdb_version'] is None:
                gui_lv = LooseVersion(el['gui_version'])
            else:
                mindsdb_lv = LooseVersion(el['mindsdb_version'])
                gui_lv = LooseVersion(el['gui_version'])
                if mindsdb_lv.vstring not in gui_versions or gui_lv > gui_versions[mindsdb_lv.vstring]:
                    gui_versions[mindsdb_lv.vstring] = gui_lv
                if max_mindsdb_lv is None or max_mindsdb_lv < mindsdb_lv:
                    max_mindsdb_lv = mindsdb_lv
            if max_gui_lv is None or max_gui_lv < gui_lv:
                max_gui_lv = gui_lv

        all_mindsdb_lv = [LooseVersion(x) for x in gui_versions.keys()]
        all_mindsdb_lv.sort()

        if current_mindsdb_lv.vstring in gui_versions:
            gui_version_lv = gui_versions[current_mindsdb_lv.vstring]
        elif current_mindsdb_lv > all_mindsdb_lv[-1]:
            gui_version_lv = max_gui_lv
        else:
            lower_versions = {key: value for key, value in gui_versions.items() if LooseVersion(key) < current_mindsdb_lv}
            if len(lower_versions) == 0:
                gui_version_lv = gui_versions[all_mindsdb_lv[0].vstring]
            else:
                all_lower_versions = [LooseVersion(x) for x in lower_versions.keys()]
                gui_version_lv = gui_versions[all_lower_versions[-1].vstring]
    except Exception as e:
        logger.error(f"Error in compatible-config.json structure: {e}")
        return False

    logger.debug(f"Last compatible frontend version: {gui_version_lv}.")
    return gui_version_lv


def get_current_gui_version() -> LooseVersion:
    logger.debug("Getting current frontend version..")
    config = Config()
    static_path = Path(config['paths']['static'])
    version_txt_path = static_path.joinpath('version.txt')

    current_gui_version = None
    if version_txt_path.is_file():
        with open(version_txt_path, 'rt') as f:
            current_gui_version = f.readline()

    current_gui_lv = (
        None if current_gui_version is None else LooseVersion(current_gui_version)
    )
    logger.debug(f"Current frontend version: {current_gui_lv}.")

    return current_gui_lv


def initialize_static():
    logger.debug("Initializing static..")
    config = Config()
    last_gui_version_lv = get_last_compatible_gui_version()
    current_gui_version_lv = get_current_gui_version()
    required_gui_version = config['gui'].get('version')

    if required_gui_version is not None:
        required_gui_version_lv = LooseVersion(required_gui_version)
        success = True
        if (
            current_gui_version_lv is None
            or required_gui_version_lv != current_gui_version_lv
        ):
            logger.debug("Updating gui..")
            success = update_static(required_gui_version_lv)
    else:
        if last_gui_version_lv is False:
            return False

        # ignore versions like '23.9.2.2'
        if current_gui_version_lv is not None and len(current_gui_version_lv.version) < 3:
            if current_gui_version_lv >= last_gui_version_lv:
                return True
        logger.debug("Updating gui..")
        success = update_static(last_gui_version_lv)

    db.session.close()
    return success


def initialize_app(config, no_studio):
    static_root = config['paths']['static']
    logger.debug(f"Static route: {static_root}")
    gui_exists = Path(static_root).joinpath('index.html').is_file()
    logger.debug(f"Does GUI already exist.. {'YES' if gui_exists else 'NO'}")
    init_static_thread = None
    if (
        no_studio is False
        and (
            config['gui']['autoupdate'] is True
            or gui_exists is False
        )
    ):
        init_static_thread = threading.Thread(target=initialize_static)
        init_static_thread.start()

    app, api = initialize_flask(config, init_static_thread, no_studio)
    Compress(app)
    initialize_interfaces(app)

    if os.path.isabs(static_root) is False:
        static_root = os.path.join(os.getcwd(), static_root)
    static_root = Path(static_root)

    @app.route('/', defaults={'path': ''}, methods=['GET'])
    @app.route('/<path:path>', methods=['GET'])
    def root_index(path):
        if path.startswith('api/'):
            return {'message': 'wrong query'}, 400
        if static_root.joinpath(path).is_file():
            return send_from_directory(static_root, path)
        else:
            return send_from_directory(static_root, 'index.html')

    protected_namespaces = [
        tab_ns,
        utils_ns,
        conf_ns,
        file_ns,
        sql_ns,
        analysis_ns,
        handlers_ns,
        tree_ns,
        projects_ns,
        databases_ns,
        views_ns,
        models_ns,
        chatbots_ns,
        skills_ns,
        agents_ns,
        jobs_ns,
        knowledge_bases_ns
    ]

    for ns in protected_namespaces:
        api.add_namespace(ns)
    api.add_namespace(default_ns)
    api.add_namespace(auth_ns)

    @api.errorhandler(Exception)
    def handle_exception(e):
        logger.error(f"http exception: {e}")
        # pass through HTTP errors
        if isinstance(e, HTTPException):
            return {"message": str(e)}, e.code, e.get_response().headers
        name = getattr(type(e), "__name__") or "Unknown error"
        return {"message": f"{name}: {str(e)}"}, 500

    @app.teardown_appcontext
    def remove_session(*args, **kwargs):
        db.session.remove()

    @app.before_request
    def before_request():
        logger.debug(f"HTTP: {request.path}")
        ctx.set_default()
        config = Config()

        # region routes where auth is required
        if (
            config['auth']['http_auth_enabled'] is True
            and any(request.path.startswith(f'/api{ns.path}') for ns in protected_namespaces)
            and check_auth() is False
        ):
            return http_error(
                403, 'Forbidden',
                'Authorization is required to complete the request'
            )
        # endregion

        company_id = request.headers.get('company-id')
        user_class = request.headers.get('user-class')

        try:
            email_confirmed = int(request.headers.get('email-confirmed', 1))
        except ValueError:
            email_confirmed = 1

        if company_id is not None:
            try:
                company_id = int(company_id)
            except Exception as e:
                logger.error(
                    f"Cloud not parse company id: {company_id} | exception: {e}"
                )
                company_id = None

        if user_class is not None:
            try:
                user_class = int(user_class)
            except Exception as e:
                logger.error(
                    f"Cloud not parse user_class: {user_class} | exception: {e}"
                )
                user_class = 0
        else:
            user_class = 0

        ctx.company_id = company_id
        ctx.user_class = user_class
        ctx.email_confirmed = email_confirmed

    # Wait for static initialization.
    if not no_studio and init_static_thread is not None:
        init_static_thread.join()

    logger.debug("Done initializing app.")
    return app


def initialize_flask(config, init_static_thread, no_studio):
    logger.debug("Initializing flask..")
    # region required for windows https://github.com/mindsdb/mindsdb/issues/2526
    mimetypes.add_type('text/css', '.css')
    mimetypes.add_type('text/javascript', '.js')
    # endregion

    kwargs = {}
    if no_studio is not True:
        static_path = os.path.join(config['paths']['static'], 'static/')
        if os.path.isabs(static_path) is False:
            static_path = os.path.join(os.getcwd(), static_path)
        kwargs["static_url_path"] = "/static"
        kwargs["static_folder"] = static_path
        logger.debug(f"Static path: {static_path}")

    app = Flask(__name__, **kwargs)
    init_metrics(app)

    app.config['SECRET_KEY'] = os.environ.get('FLASK_SECRET_KEY', secrets.token_hex(32))
    app.config['SESSION_COOKIE_NAME'] = 'session'
    app.config['PERMANENT_SESSION_LIFETIME'] = datetime.timedelta(days=31)
    app.config['SEND_FILE_MAX_AGE_DEFAULT'] = 60
    app.config['SWAGGER_HOST'] = 'http://localhost:8000/mindsdb'
    app.json_encoder = CustomJSONEncoder

    authorizations = {
        'apikey': {
            'type': 'session',
            'in': 'query',
            'name': 'session'
        }
    }

    logger.debug("Creating swagger API..")
    api = Swagger_Api(
        app,
        authorizations=authorizations,
        security=['apikey'],
        url_prefix=':8000',
        prefix='/api',
        doc='/doc/'
    )

    api.representations['application/json'] = custom_output_json

    port = config['api']['http']['port']
    host = config['api']['http']['host']

    # NOTE rewrite it, that hotfix to see GUI link
    if not no_studio:
        if host in ("", "0.0.0.0"):
            url = f"http://127.0.0.1:{port}/"
        else:
            url = f"http://{host}:{port}/"
        logger.info(f" - GUI available at {url}")

        pid = os.getpid()
        x = threading.Thread(target=_open_webbrowser, args=(url, pid, port, init_static_thread, config['paths']['static']), daemon=True)
        x.start()

    return app, api


def initialize_interfaces(app):
    app.integration_controller = integration_controller
    app.database_controller = DatabaseController()
    app.file_controller = FileController()
    config = Config()
    app.config_obj = config


def _open_webbrowser(url: str, pid: int, port: int, init_static_thread, static_folder):
    """Open webbrowser with url when http service is started.

    If some error then do nothing.
    """
    if init_static_thread is not None:
        init_static_thread.join()
    inject_telemetry_to_static(static_folder)
    try:
        is_http_active = wait_func_is_true(
            func=is_pid_listen_port, timeout=15, pid=pid, port=port
        )
        if is_http_active:
            webbrowser.open(url)
    except Exception as e:
        logger.error(f"Failed to open {url} in webbrowser with exception {e}")
        logger.error(traceback.format_exc())
    db.session.close()



import requests


class AQIClient:
    def __init__(self, api_key):
        self.api_key = api_key
        self.params = {"token": api_key}
        self.base_endpoint = "https://api.waqi.info/feed"

    def make_request(self, url):
        resp = requests.get(url, params=self.params)
        res = resp.json()
        content = {}
        if res["status"] == "ok":
            content = {'content': resp.json(), 'code': 200}
        else:
            content = {'content': resp.json(), 'code': 404}
        return content

    def air_quality_city(self, city):
        url = f'{self.base_endpoint}/{city}/'
        return self.make_request(url)

    def air_quality_lat_lng(self, lat, lng):
        url = f'{self.base_endpoint}/geo:{lat};{lng}/'
        return self.make_request(url)

    def air_quality_user_location(self):
        url = f'{self.base_endpoint}/here/'
        return self.make_request(url)



from http import HTTPStatus

from flask import request
from flask_restx import Resource

from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.api.http.utils import http_error
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.interfaces.chatbot.chatbot_controller import ChatBotController
from mindsdb.interfaces.model.functions import PredictorRecordNotFound
from mindsdb.interfaces.storage.db import Predictor


def create_chatbot(project_name, name, chatbot):
    if name is None:
        return http_error(
            HTTPStatus.BAD_REQUEST,
            'Missing field',
            'Missing "name" field for chatbot'
        )

    model_name = chatbot.get('model_name', None)
    agent_name = chatbot.get('agent_name', None)
    if model_name is None and agent_name is None:
        return http_error(
            HTTPStatus.BAD_REQUEST,
            'Missing field',
            'Must include either "model_name" or "agent_name" field for chatbot'
        )

    session_controller = SessionController()

    if 'database_id' not in chatbot:
        if 'db_params' in chatbot and 'db_engine' in chatbot:
            db_name = chatbot['name'] + '_db'

            # try to drop
            existing_db = session_controller.integration_controller.get(db_name)
            if existing_db:
                # drop
                session_controller.integration_controller.delete(db_name)

            database_id = session_controller.integration_controller.add(db_name, chatbot['db_engine'],
                                                                        chatbot['db_params'])

        else:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing field',
                'Missing "database_id" or ("db_engine" and "database_param") fields for chatbot'
            )
    else:
        database_id = chatbot.get('database_id', None)

    is_running = chatbot.get('is_running', False)
    params = chatbot.get('params', {})

    chatbot_controller = ChatBotController()

    # Chatbot can't already exist.
    # TODO all checks should be inside of controller

    try:
        existing_chatbot = chatbot_controller.get_chatbot(name, project_name=project_name)
    except ValueError:
        # Project must exist.
        return http_error(
            HTTPStatus.NOT_FOUND,
            'Project not found',
            f'Project with name {project_name} does not exist')
    if existing_chatbot is not None:
        return http_error(
            HTTPStatus.CONFLICT,
            'Chatbot already exists',
            f'Chatbot with name {name} already exists. Please use a different name'
        )

    # Model and agent need to exist.
    if agent_name is not None:
        agent = session_controller.agents_controller.get_agent(agent_name, project_name)
        if agent is None:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Agent not found',
                f'Agent with name {agent_name} not found'
            )
        model_name = agent.model_name

    model_name_no_version, version = Predictor.get_name_and_version(model_name)
    try:
        session_controller.model_controller.get_model(model_name_no_version, version=version, project_name=project_name)
    except PredictorRecordNotFound:
        return http_error(
            HTTPStatus.NOT_FOUND,
            'Model not found',
            f'Model with name {model_name} not found')

    created_chatbot = chatbot_controller.add_chatbot(
        name,
        project_name,
        model_name=model_name,
        agent_name=agent_name,
        database_id=database_id,
        is_running=is_running,
        params=params
    )
    return created_chatbot.as_dict(), HTTPStatus.CREATED


@ns_conf.route('/<project_name>/chatbots')
class ChatBotsResource(Resource):
    @ns_conf.doc('list_chatbots')
    @api_endpoint_metrics('GET', '/chatbots')
    def get(self, project_name):
        ''' List all chatbots '''
        chatbot_controller = ChatBotController()
        try:
            all_bots = chatbot_controller.get_chatbots(project_name)
        except ValueError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist')
        return all_bots

    @ns_conf.doc('create_chatbot')
    @api_endpoint_metrics('POST', '/chatbots')
    def post(self, project_name):
        '''Create a chatbot'''

        # Check for required parameters.
        if 'chatbot' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "chatbot" parameter in POST body'
            )

        chatbot = request.json['chatbot']

        name = chatbot.get('name')
        return create_chatbot(project_name, name, chatbot)


@ns_conf.route('/<project_name>/chatbots/<chatbot_name>')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('chatbot_name', 'Name of the chatbot')
class ChatBotResource(Resource):
    @ns_conf.doc('get_chatbot')
    @api_endpoint_metrics('GET', '/chatbots/chatbot')
    def get(self, project_name, chatbot_name):
        '''Gets a chatbot by name'''
        chatbot_controller = ChatBotController()
        try:
            existing_chatbot = chatbot_controller.get_chatbot(chatbot_name, project_name=project_name)
            if existing_chatbot is None:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Chatbot not found',
                    f'Chatbot with name {chatbot_name} does not exist'
                )
            return existing_chatbot
        except ValueError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

    @ns_conf.doc('update_chatbot')
    @api_endpoint_metrics('PUT', '/chatbots/chatbot')
    def put(self, project_name, chatbot_name):
        '''Updates a chatbot by name, creating one if it doesn't exist'''

        # Check for required parameters.
        if 'chatbot' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "chatbot" parameter in POST body'
            )
        chatbot_controller = ChatBotController()

        try:
            existing_chatbot = chatbot_controller.get_chatbot(chatbot_name, project_name=project_name)
        except ValueError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        chatbot = request.json['chatbot']
        name = chatbot.get('name', None)
        agent_name = chatbot.get('agent_name', None)
        model_name = chatbot.get('model_name', None)
        database_id = chatbot.get('database_id', None)
        is_running = chatbot.get('is_running', None)
        params = chatbot.get('params', None)

        # Model needs to exist.
        session = SessionController()
        if model_name is not None:
            model_name_no_version, version = Predictor.get_name_and_version(model_name)
            try:
                session.model_controller.get_model(model_name_no_version, version=version, project_name=project_name)
            except PredictorRecordNotFound:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Model not found',
                    f'Model with name {model_name} not found')

        # Agent needs to exist.
        if agent_name is not None:
            agent = session.agents_controller.get_agent(agent_name, project_name)
            if agent is None:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Agent not found',
                    f'Agent with name {agent_name} not found')

        # Chatbot must not exist with new name.
        if name is not None and name != chatbot_name:
            chatbot_with_new_name = chatbot_controller.get_chatbot(name, project_name=project_name)
            if chatbot_with_new_name is not None:
                return http_error(
                    HTTPStatus.CONFLICT,
                    'Chatbot already exists',
                    f'Chatbot with name {name} already exists. Please choose a different one.'
                )

        if existing_chatbot is None:
            # Create
            return create_chatbot(project_name, name, chatbot)

        if 'db_params' in chatbot and 'db_engine' in chatbot:
            session_controller = SessionController()

            db_name = chatbot['name'] + '_db'

            # try to drop
            existing_db = session_controller.integration_controller.get(db_name)
            if existing_db:
                # drop
                session_controller.integration_controller.delete(db_name)

            database_id = session_controller.integration_controller.add(db_name, chatbot['db_engine'],
                                                                        chatbot['db_params'])
        # Update
        updated_chatbot = chatbot_controller.update_chatbot(
            chatbot_name,
            project_name=project_name,
            name=name,
            model_name=model_name,
            agent_name=agent_name,
            database_id=database_id,
            is_running=is_running,
            params=params
        )
        return updated_chatbot.as_dict()

    @ns_conf.doc('delete_chatbot')
    @api_endpoint_metrics('DELETE', '/chatbots/chatbot')
    def delete(self, project_name, chatbot_name):
        '''Deletes a chatbot by name'''
        chatbot_controller = ChatBotController()
        try:
            existing_chatbot = chatbot_controller.get_chatbot(chatbot_name, project_name=project_name)
            if existing_chatbot is None:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Chatbot not found',
                    f'Chatbot with name {chatbot_name} does not exist'
                )
        except ValueError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        chatbot_controller.delete_chatbot(chatbot_name, project_name=project_name)
        return '', HTTPStatus.NO_CONTENT



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __version__ as version, __description__ as description

try:
    from .timegpt_handler import TimeGPTHandler as Handler
    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = "TimeGPT"
name = "timegpt"
type = HANDLER_TYPE.ML
icon_path = "icon.svg"
permanent = False

__all__ = ["Handler", "version", "name", "type", "title", "description", "import_error", "icon_path"]



import os
from llama_hub.github_repo import GithubRepositoryReader


def _get_github_token(args, connection_args):
    """
    API_KEY preference order:
        1. provided at model creation
        2. provided at engine creation
        3. GITHUB_TOKEN env variable

    Note: method is not case-sensitive.
    """
    key = "GITHUB_TOKEN"
    for k in key, key.lower():
        if args.get(k):
            return args[k]

        connection_args = connection_args
        if connection_args.get(k):
            return connection_args[k]

        api_key = os.getenv(k)
        if os.environ.get(k):
            return api_key

    return None


def _get_filter_file_extensions(args):
    """
    Returns file extensions to filter, if Filter type is EXCLUDE the file extensions will be excluded
    from the knowledge source, if Filter type is INCLUDE the file extensions will be included in the
    knowledge source.
    """
    # filter_file_extensions is not provided
    if "filter_file_extensions" not in args:
        return None

    # if filter_type is provided with EXCLUDE or INCLUDE
    if args["filter_file_extensions"] and args["filter_type"].upper() == "INCLUDE":
        filter_file_extensions = args["filter_file_extensions"]
        return (filter_file_extensions, GithubRepositoryReader.FilterType.INCLUDE)
    else:
        filter_file_extensions = args["filter_file_extensions"]
        return (filter_file_extensions, GithubRepositoryReader.FilterType.EXCLUDE)


def _get_filter_directories(args):
    """
    Returns directories to filter, if Filter type is EXCLUDE the directories will be excluded
    from the knowledge source, if Filter type is INCLUDE the directories will be included in the
    knowledge source.
    """
    # filter_directories is not provided
    if "filter_directories" not in args:
        return None

    # if filter_type is provided with EXCLUDE or INCLUDE
    if args["filter_directories"] and args["filter_type"].upper() == "INCLUDE":
        filter_directories = args["filter_directories"]
        return (filter_directories, GithubRepositoryReader.FilterType.INCLUDE)
    else:
        filter_directories = args["filter_directories"]
        return (filter_directories, GithubRepositoryReader.FilterType.EXCLUDE)



from dataclasses import dataclass
from typing import Callable, List, Union

from langchain_core.documents import Document
from langchain_text_splitters import MarkdownHeaderTextSplitter, HTMLHeaderTextSplitter, RecursiveCharacterTextSplitter

from mindsdb.utilities import log

DEFAULT_CHUNK_SIZE = 1000
DEFAULT_CHUNK_OVERLAP = 50
DEFAULT_MARKDOWN_HEADERS_TO_SPLIT_ON = [
    ('#', 'Header 1'),
    ('##', 'Header 2'),
    ('###', 'Header 3')
]
DEFAULT_HTML_HEADERS_TO_SPLIT_ON = [
    ('h1', 'Header 1'),
    ('h2', 'Header 2'),
    ('h3', 'Header 3'),
    ('h4', 'Header 4')
]
logger = log.getLogger(__name__)


@dataclass
class FileSplitterConfig:
    '''Represents configuration needed to split a file into chunks for retrieval.'''
    # Target chunk size in characters. Not all splitters will adhere exactly to this (it's more of a guideline)
    chunk_size: int = DEFAULT_CHUNK_SIZE
    # How many characters each chunk should overlap. Not all splitters will adhere exactly to this (it's more of a guideline)
    chunk_overlap: int = DEFAULT_CHUNK_OVERLAP
    # Default recursive splitter to use for text files, or unsupported files
    recursive_splitter: RecursiveCharacterTextSplitter = RecursiveCharacterTextSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)
    # Splitter to use for MD splitting
    markdown_splitter: MarkdownHeaderTextSplitter = MarkdownHeaderTextSplitter(headers_to_split_on=DEFAULT_MARKDOWN_HEADERS_TO_SPLIT_ON)
    # Splitter to use for HTML splitting
    html_splitter: HTMLHeaderTextSplitter = HTMLHeaderTextSplitter(headers_to_split_on=DEFAULT_HTML_HEADERS_TO_SPLIT_ON)


class FileSplitter:
    '''Splits Documents that represent various file types into chunks for retrieval.'''
    def __init__(self, config: FileSplitterConfig):
        '''
        Args:
            config (FileSplitterConfig): Configuration for the file splitter.
        '''
        self.config = config
        self._extension_map = {
            '.pdf': self._recursive_splitter_fn,
            '.md': self._markdown_splitter_fn,
            '.html': self._html_splitter_fn
        }
        self.default_splitter = self._recursive_splitter_fn

    def _split_func_by_extension(self, extension) -> Union[Callable, HTMLHeaderTextSplitter, MarkdownHeaderTextSplitter]:
        return self._extension_map.get(extension, self.default_splitter)()

    def split_documents(self, documents: List[Document], default_failover: bool = True) -> List[Document]:
        '''Splits a list of documents representing files using the appropriate splitting & chunking strategies

        Args:
            documents (List[Document]): List of documents representing files to split.
            default_failover (bool, optional): Whether to use the default splitter as a fallback if the file type is not supported. Defaults to True.

        Returns:
            List[Document]: List of documents representing the split files.
        '''
        split_documents = []
        document: Document
        for document in documents:
            extension = document.metadata['extension']
            split_func = self._split_func_by_extension(extension=extension)
            try:
                split_documents += split_func(document.page_content)
            except Exception as e:
                logger.error(f'Error splitting document with extension {extension}: {str(e)}')
                if not default_failover:
                    raise ValueError(f'Error splitting document with extension {extension}') from e
                # Try default splitter as a failover, if enabled.
                split_func = self._split_func_by_extension(extension=None)
                split_documents += split_func(document.page_content)
        return split_documents

    def _markdown_splitter_fn(self) -> Callable:
        return self.config.markdown_splitter.split_text

    def _html_splitter_fn(self) -> Callable:
        return self.config.html_splitter.split_text

    def _recursive_splitter_fn(self) -> Callable:
        # Recursive splitter is a TextSplitter where split_text returns List[str].
        def recursive_split(content: str) -> List[Document]:
            split_content = self.config.recursive_splitter.split_text(content)
            return [Document(page_content=c) for c in split_content]
        return recursive_split



from typing import List
import uuid

from langchain.retrievers.multi_vector import MultiVectorRetriever as LangChainMultiVectorRetriever
from langchain_core.documents import Document
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI

from mindsdb.integrations.utilities.rag.retrievers.base import BaseRetriever
from mindsdb.integrations.utilities.rag.settings import DEFAULT_LLM_MODEL, \
    MultiVectorRetrieverMode, RAGPipelineModel
from mindsdb.integrations.utilities.rag.vector_store import VectorStoreOperator


class MultiVectorRetriever(BaseRetriever):
    """
    MultiVectorRetriever stores multiple vectors per document.
    """

    def __init__(self, config: RAGPipelineModel):
        self.vectorstore = config.vector_store
        self.parent_store = config.parent_store
        self.id_key = config.id_key
        self.documents = config.documents
        self.text_splitter = config.text_splitter
        self.embeddings_model = config.embeddings_model
        self.max_concurrency = config.max_concurrency
        self.mode = config.multi_retriever_mode

    def _generate_id_and_split_document(self, doc: Document) -> tuple[str, list[Document]]:
        """
        Generate a unique id for the document and split it into sub-documents.
        :param doc:
        :return:
        """
        doc_id = str(uuid.uuid4())
        sub_docs = self.text_splitter.split_documents([doc])
        for sub_doc in sub_docs:
            sub_doc.metadata[self.id_key] = doc_id
        return doc_id, sub_docs

    def _split_documents(self) -> tuple[list[Document], list[str]]:
        """
        Split the documents into sub-documents and generate unique ids for each document.
        :return:
        """
        split_info = list(map(self._generate_id_and_split_document, self.documents))
        doc_ids, split_docs_lists = zip(*split_info)
        split_docs = [doc for sublist in split_docs_lists for doc in sublist]
        return split_docs, list(doc_ids)

    def _create_retriever_and_vs_operator(self, docs: List[Document]) \
            -> tuple[LangChainMultiVectorRetriever, VectorStoreOperator]:
        vstore_operator = VectorStoreOperator(
            vector_store=self.vectorstore,
            documents=docs,
            embeddings_model=self.embeddings_model,
        )
        retriever = LangChainMultiVectorRetriever(
            vectorstore=vstore_operator.vector_store,
            byte_store=self.parent_store,
            id_key=self.id_key
        )
        return retriever, vstore_operator

    def _get_document_summaries(self) -> List[str]:
        chain = (
                {"doc": lambda x: x.page_content}  # noqa: E126, E122
                | ChatPromptTemplate.from_template("Summarize the following document:\n\n{doc}")
                | ChatOpenAI(max_retries=0, model_name=DEFAULT_LLM_MODEL)
                | StrOutputParser()
        )
        return chain.batch(self.documents, {"max_concurrency": self.max_concurrency})

    def as_runnable(self) -> BaseRetriever:
        if self.mode in {MultiVectorRetrieverMode.SPLIT, MultiVectorRetrieverMode.BOTH}:
            split_docs, doc_ids = self._split_documents()
            retriever, vstore_operator = self._create_retriever_and_vs_operator(split_docs)
            summaries = self._get_document_summaries()
            summary_docs = [
                Document(page_content=s, metadata={self.id_key: doc_ids[i]})
                for i, s in enumerate(summaries)
            ]
            vstore_operator.add_documents(summary_docs)
            retriever.docstore.mset(list(zip(doc_ids, self.documents)))
            return retriever

        elif self.mode == MultiVectorRetrieverMode.SUMMARIZE:
            summaries = self._get_document_summaries()
            doc_ids = [str(uuid.uuid4()) for _ in self.documents]
            summary_docs = [
                Document(page_content=s, metadata={self.id_key: doc_ids[i]})
                for i, s in enumerate(summaries)
            ]
            retriever, vstore_operator = self._create_retriever_and_vs_operator(summary_docs)
            retriever.docstore.mset(list(zip(doc_ids, self.documents)))
            return retriever

        else:
            raise ValueError(f"Invalid mode: {self.mode}")



"""drop-semaphor

Revision ID: 2958416fbe75
Revises: 9461892bd889
Create Date: 2024-04-25 18:30:54.051212

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = '2958416fbe75'
down_revision = '9461892bd889'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_table('semaphor')


def downgrade():
    op.create_table(
        'semaphor',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.Column('updated_at', sa.DATETIME(), nullable=True),
        sa.Column('created_at', sa.DATETIME(), nullable=True),
        sa.Column('entity_type', sa.VARCHAR(), nullable=True),
        sa.Column('entity_id', sa.INTEGER(), nullable=True),
        sa.Column('action', sa.VARCHAR(), nullable=True),
        sa.Column('company_id', sa.INTEGER(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('entity_type', 'entity_id', name='uniq_const')
    )



import paypalrestsdk

from mindsdb.integrations.handlers.paypal_handler.paypal_tables import (
    InvoicesTable,
    PaymentsTable,
    SubscriptionsTable,
    OrdersTable,
    PayoutsTable
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
  HandlerStatusResponse as StatusResponse,
)

from mindsdb.utilities import log
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb_sql import parse_sql
from collections import OrderedDict

logger = log.getLogger(__name__)


class PayPalHandler(APIHandler):
    """
    The PayPal handler implementation.
    """

    name = 'paypal'

    def __init__(self, name: str, **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        payments_data = PaymentsTable(self)
        self._register_table("payments", payments_data)

        invoices_data = InvoicesTable(self)
        self._register_table("invoices", invoices_data)

        subscriptions_data = SubscriptionsTable(self)
        self._register_table("subscriptions", subscriptions_data)

        orders_data = OrdersTable(self)
        self._register_table("orders", orders_data)

        payouts_data = PayoutsTable(self)
        self._register_table("payouts", payouts_data)

    def connect(self):
        """
        Set up the connection required by the handler.
        Returns
        -------
        StatusResponse
            connection object
        """
        if self.is_connected is True:
            return self.connection

        self.connection = paypalrestsdk.Api(
            {
                "mode": self.connection_data['mode'],
                "client_id": self.connection_data['client_id'],
                "client_secret": self.connection_data['client_secret'],
            }
        )

        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)

        try:
            connection = self.connect()
            connection.get_access_token()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to PayPal!')
            response.error_message = str(e)

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)


connection_args = OrderedDict(
    mode={
        "type": ARG_TYPE.STR,
        "description": "Environment mode of the app",
        "required": True,
        "label": "MODE",
    },
    client_id={
        "type": ARG_TYPE.PWD,
        "description": "Client id of the App",
        "required": True,
        "label": "Client ID",
    },
    client_secret={
        "type": ARG_TYPE.STR,
        "description": "Client secret of the App",
        "required": True,
        "label": "Client Secret",
    },
)

connection_args_example = OrderedDict(
    mode="sandbox",
    client_id="xxxx-xxxx-xxxx-xxxx",
    client_secret="<secret>",
)



from textwrap import dedent
from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The IP address/host name of the SAP HANA instance host.'
    },
    port={
        'type': ARG_TYPE.STR,
        'description': 'The port number of the SAP HANA instance.'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the user name.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'Specifies the password for the user.',
        'secret': True
    },
    schema={
        'type': ARG_TYPE.STR,
        'description': 'Sets the current schema, which is used for identifiers without a schema.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the name of the database to connect to. (Not used for SAP HANA Cloud)'
    },
    autocommit={
        'type': ARG_TYPE.BOOL,
        'description': 'Sets the autocommit mode for the connection.'
    },
    properties={
        'type': ARG_TYPE.STR,
        'description': 'Additional dictionary with special properties of the connection.'
    },
    encrypt={
        'type': ARG_TYPE.BOOL,
        'description': 'Enables or disables TLS encryption.'
    },
    sslHostNameInCertificate={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the host name used to verify server\'s identity.'
    },
    sslValidateCertificate={
        'type': ARG_TYPE.BOOL,
        'description': 'Specifies whether to validate the server\'s certificate.'
    },
    sslCryptoProvider={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the cryptographic library provider used for TLS communication.'
    },
    sslTrustStore={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the path to a trust store file that contains the server\'s public certificates.'
    },
    sslKeyStore={
        'type': ARG_TYPE.STR,
        'description': 'Specifies the path to the keystore file that contains the client\'s identity.'
    },
    cseKeyStorePassword={
        'type': ARG_TYPE.STR,
        'description': 'Provides the password for the local key store.'
    },
    sslSNIHostname={
        'type': ARG_TYPE.STR,
        'description': dedent("""Specifies the name of the host that is attempting to connect at the start of
        the TLS handshaking process.""")
    },
    sslSNIRequest={
        'type': ARG_TYPE.BOOL,
        'description': 'Specifies whether SNI requests are enabled for TLS connections: TRUE/FALSE.'
    },
    siteType={
        'type': ARG_TYPE.STR,
        'description': dedent("""Specifies whether the connection is made to either the PRIMARY or SECONDARY
        site in an Active/Active (read enabled) system.""")
    },
    splitBatchCommands={
        'type': ARG_TYPE.BOOL,
        'description': 'Allows split and parallel execution of batch commands on partitioned tables.'
    },
    routeDirectExecute={
        'type': ARG_TYPE.BOOL,
        'description': dedent("""Converts direct execute into prepare and execute (routed execute) if the
        number of index servers is more than one and if statement routing is enabled.""")
    },
    secondarySessionFallback={
        'type': ARG_TYPE.BOOL,
        'description': dedent("""Forces the ongoing transaction on a non-anchor connection to fall back
        to the anchor/primary connection if this connection is dropped by the network or server.""")
    }
)

connection_args_example = OrderedDict(
    host='<uuid>.hana.trial-us10.hanacloud.ondemand.com',
    port=30013,
    user='DBADMIN',
    password='password',
    schema='MINDSDB',
)



import os
from typing import Optional, Dict

import openai
import pandas as pd
import llama_index

from llama_index.llms.openai import OpenAI
from llama_index.readers.schema.base import Document
from llama_index.readers import SimpleWebPageReader
from llama_index.prompts import PromptTemplate
from llama_index import ServiceContext, StorageContext, load_index_from_storage
from llama_index import LLMPredictor, OpenAIEmbedding
from llama_index.indices.vector_store.base import VectorStore

from llama_hub.github_repo import GithubClient, GithubRepositoryReader
from llama_hub.youtube_transcript import YoutubeTranscriptReader, is_youtube_video

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities.config import Config
from mindsdb.utilities.security import is_private_url
from mindsdb.integrations.handlers.llama_index_handler import config
from mindsdb.integrations.handlers.llama_index_handler.github_loader_helper import (
    _get_github_token,
    _get_filter_file_extensions,
    _get_filter_directories,
)
from mindsdb.integrations.utilities.handler_utils import get_api_key


def _validate_prompt_template(prompt_template: str):
    if "{context_str}" not in prompt_template or "{query_str}" not in prompt_template:
        raise Exception(
            "Provided prompt template is invalid, missing `{context_str}`, `{query_str}`. Please ensure both placeholders are present and try again."
        )  # noqa


class LlamaIndexHandler(BaseMLEngine):
    """Integration with the LlamaIndex data framework for LLM applications."""

    name = "llama_index"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True
        self.default_index_class = "GPTVectorStoreIndex"
        self.supported_index_class = ["GPTVectorStoreIndex", "VectorStoreIndex"]
        self.default_reader = "DFReader"
        self.supported_reader = [
            "DFReader",
            "SimpleWebPageReader",
            "GithubRepositoryReader",
            "YoutubeTranscriptReader",
        ]

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        reader = args["using"].get("reader", "DFReader")

        if reader not in config.data_loaders:
            raise Exception(
                f"Invalid reader argument. Please use one of {config.data_loaders.keys()}"
            )
        config_dict = config.data_loaders[reader]

        missing_keys = [key for key in config_dict if key not in args["using"]]
        if missing_keys:
            raise Exception(f"{reader} requires {missing_keys} arguments")

        if "prompt_template" in args["using"]:
            _validate_prompt_template(args["using"]["prompt_template"])

        if args["using"].get("mode") == "conversational":
            for param in ("user_column", "assistant_column"):
                if param not in args["using"]:
                    raise Exception(f"Conversational mode requires {param} parameter")

    def create(
        self,
        target: str,
        df: Optional[pd.DataFrame] = None,
        args: Optional[Dict] = None,
    ) -> None:
        if "using" not in args:
            raise Exception(
                "LlamaIndex engine requires a USING clause! Refer to its documentation for more details."
            )

        if "index_class" not in args["using"]:
            args["using"]["index_class"] = self.default_index_class
        elif args["using"]["index_class"] not in self.supported_index_class:
            raise Exception(
                f"Invalid index class argument. Please use one of {self.supported_index_class}"
            )

        if "reader" not in args["using"]:
            args["using"]["reader"] = self.default_reader
        elif args["using"]["reader"] not in self.supported_reader:
            raise Exception(
                f"Invalid operation mode. Please use one of {self.supported_reader}"
            )

        # workaround to create llama model without input data
        if df is None or df.empty:
            df = pd.DataFrame([{"text": ""}])

        if args["using"]["reader"] == "DFReader":
            dstrs = df.apply(
                lambda x: ", ".join(
                    [f"{col}: {str(entry)}" for col, entry in zip(df.columns, x)]
                ),
                axis=1,
            )
            reader = list(map(lambda x: Document(text=x), dstrs.tolist()))

        elif args["using"]["reader"] == "SimpleWebPageReader":
            url = args["using"]["source_url_link"]
            config = Config()
            is_cloud = config.get("cloud", False)
            if is_cloud and is_private_url(url):
                raise Exception(f"URL is private: {url}")

            reader = SimpleWebPageReader(html_to_text=True).load_data([url])

        elif args["using"]["reader"] == "GithubRepositoryReader":
            engine_storage = self.engine_storage
            key = "GITHUB_TOKEN"
            github_token = get_api_key(
                key, args["using"], engine_storage, strict=False
            )
            if github_token is None:
                github_token = get_api_key(
                    key.lower(),
                    args["using"],
                    engine_storage,
                    strict=True,
                )

            github_client = GithubClient(github_token)
            owner = args["using"]["owner"]
            repo = args["using"]["repo"]
            filter_file_extensions = _get_filter_file_extensions(args["using"])
            filter_directories = _get_filter_directories(args["using"])

            reader = GithubRepositoryReader(
                github_client,
                owner=owner,
                repo=repo,
                verbose=True,
                filter_file_extensions=filter_file_extensions,
                filter_directories=filter_directories,
            ).load_data(branch=args["using"].get("branch", "main"))

        elif args["using"]["reader"] == "YoutubeTranscriptReader":
            ytlinks = args["using"]["ytlinks"]
            for link in ytlinks:
                if not is_youtube_video(link):
                    raise Exception(f"Invalid youtube link: {link}")
            reader = YoutubeTranscriptReader().load_data(ytlinks)

        else:
            raise Exception(
                f"Invalid operation mode. Please use one of {self.supported_reader}."
            )

        self.model_storage.json_set("args", args)
        index = self._setup_index(reader)
        path = self.model_storage.folder_get("context")
        index.storage_context.persist(persist_dir=path)
        self.model_storage.folder_sync("context")

    def update(self, args) -> None:
        prompt_template = args["using"].get(
            "prompt_template", args.get("prompt_template", None)
        )
        if prompt_template is not None:
            _validate_prompt_template(prompt_template)
        args_cur = self.model_storage.json_get("args")
        args_cur["using"].update(args["using"])

        # check new set of arguments
        self.create_validation(None, args_cur)

        self.model_storage.json_set("args", args_cur)

    def predict(
        self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None
    ) -> pd.DataFrame:
        pred_args = args["predict_params"] if args else {}

        args = self.model_storage.json_get("args")
        engine_kwargs = {}

        if args["using"].get("mode") == "conversational":
            user_column = args["using"]["user_column"]
            assistant_column = args["using"]["assistant_column"]

            messages = []
            for row in df[:-1].to_dict("records"):
                messages.append(f"user: {row[user_column]}")
                messages.append(f"assistant: {row[assistant_column]}")

            conversation = "\n".join(messages)

            questions = [df.iloc[-1][user_column]]

            if "prompt" in pred_args and pred_args["prompt"] is not None:
                user_prompt = pred_args["prompt"]
            else:
                user_prompt = args["using"].get("prompt", "")

            prompt_template = (
                f"{user_prompt}\n"
                f"---------------------\n"
                f"We have provided context information below. \n"
                f"{{context_str}}\n"
                f"---------------------\n"
                f"This is previous conversation history:\n"
                f"{conversation}\n"
                f"---------------------\n"
                f"Given this information, please answer the question: {{query_str}}"
            )

            engine_kwargs["text_qa_template"] = PromptTemplate(prompt_template)

        else:
            input_column = args["using"].get("input_column", None)

            prompt_template = args["using"].get(
                "prompt_template", args.get("prompt_template", None)
            )
            if prompt_template is not None:
                _validate_prompt_template(prompt_template)
                engine_kwargs["text_qa_template"] = PromptTemplate(prompt_template)

            if input_column is None:
                raise Exception(
                    f"`input_column` must be provided at model creation time or through USING clause when predicting. Please try again."
                )  # noqa

            if input_column not in df.columns:
                raise Exception(
                    f'Column "{input_column}" not found in input data! Please try again.'
                )

            questions = df[input_column]

        index_path = self.model_storage.folder_get("context")
        storage_context = StorageContext.from_defaults(persist_dir=index_path)
        service_context = self._get_service_context()
        index = load_index_from_storage(
            storage_context, service_context=service_context
        )
        query_engine = index.as_query_engine(**engine_kwargs)

        results = []

        for question in questions:
            query_results = query_engine.query(
                question
            )  # TODO: provide extra_info in explain_target col
            results.append(query_results.response)

        result_df = pd.DataFrame(
            {"question": questions, args["target"]: results}
        )  # result_df['answer'].tolist()
        return result_df

    def _get_service_context(self):
        args = self.model_storage.json_get("args")
        engine_storage = self.engine_storage

        openai_api_key = get_api_key('openai', args["using"], engine_storage, strict=True)
        llm_kwargs = {"openai_api_key": openai_api_key}

        if "temperature" in args["using"]:
            llm_kwargs["temperature"] = args["using"]["temperature"]
        if "model_name" in args["using"]:
            llm_kwargs["model_name"] = args["using"]["model_name"]
        if "max_tokens" in args["using"]:
            llm_kwargs["max_tokens"] = args["using"]["max_tokens"]

        llm = OpenAI(**llm_kwargs)  # TODO: all usual params should go here
        embed_model = OpenAIEmbedding(api_key=openai_api_key)
        service_context = ServiceContext.from_defaults(
            llm=llm,
            embed_model=embed_model
        )
        return service_context

    def _setup_index(self, documents):
        args = self.model_storage.json_get("args")
        indexer: VectorStore = getattr(llama_index, args["using"]["index_class"])
        index = indexer.from_documents(
            documents, service_context=self._get_service_context()
        )

        return index




from mindsdb_sql.parser.ast import Identifier, Select, BinaryOperation, Constant, OrderBy

from mindsdb.interfaces.storage import db


from .types import ChatBotMessage


class BaseMemory:
    '''
    base class to work with chatbot memory
    '''
    MAX_DEPTH = 100

    def __init__(self, chat_task, chat_params):
        # in memory yet
        self._modes = {}
        self._hide_history_before = {}
        self._cache = {}
        self.chat_params = chat_params
        self.chat_task = chat_task

    def get_chat(self, chat_id):
        return ChatMemory(self, chat_id)

    def hide_history(self, chat_id, left_count):
        '''
        set date to start hiding messages
        '''
        history = self.get_chat_history(chat_id)
        if left_count > len(history) - 1:
            left_count = len(history) - 1
        sent_at = history[-left_count].sent_at

        self._hide_history_before[chat_id] = sent_at

    def _apply_hiding(self, chat_id, history):
        '''
        hide messages from history
        '''
        before = self._hide_history_before.get(chat_id)

        if before is None:
            return history

        return [
            msg
            for msg in history
            if msg.sent_at >= before
        ]

    def get_mode(self, chat_id):
        return self._modes.get(chat_id)

    def set_mode(self, chat_id, mode):
        self._modes[chat_id] = mode

    def add_to_history(self, chat_id, chat_message):

        self._add_to_history(chat_id, chat_message)
        if chat_id in self._cache:
            del self._cache[chat_id]

    def get_chat_history(self, chat_id, cached=True):
        if cached and chat_id in self._cache:
            history = self._cache[chat_id]
        else:
            history = self._get_chat_history(chat_id)
            self._cache[chat_id] = history

        history = self._apply_hiding(chat_id, history)
        return history

    def _add_to_history(self, chat_id, chat_message):
        raise NotImplementedError

    def _get_chat_history(self, chat_id):
        raise NotImplementedError


class HandlerMemory(BaseMemory):
    '''
    Uses handler's database to store and retrieve messages
    '''

    def _add_to_history(self, chat_id, chat_message):
        # do nothing. sent message will be stored by handler db
        pass

    def _get_chat_history(self, chat_id):
        t_params = self.chat_params['chat_table']

        text_col = t_params['text_col']
        username_col = t_params['username_col']
        time_col = t_params['time_col']

        ast_query = Select(
            targets=[Identifier(text_col),
                     Identifier(username_col),
                     Identifier(time_col)],
            from_table=Identifier(t_params['name']),
            where=BinaryOperation(
                op='=',
                args=[
                    Identifier(t_params['chat_id_col']),
                    Constant(chat_id)
                ]
            ),
            order_by=[OrderBy(Identifier(time_col))],
            limit=Constant(self.MAX_DEPTH),
        )

        resp = self.chat_task.chat_handler.query(ast_query)
        if resp.data_frame is None:
            return

        df = resp.data_frame

        # get last messages
        df = df.iloc[-self.MAX_DEPTH:]

        result = []
        for _, rec in df.iterrows():
            chatbot_message = ChatBotMessage(
                ChatBotMessage.Type.DIRECT,
                rec[text_col],
                user=rec[username_col],
                sent_at=rec[time_col]
            )
            result.append(chatbot_message)

        return result


class DBMemory(BaseMemory):
    '''
    uses mindsdb database to store messages
    '''

    def _add_to_history(self, chat_id, message):

        chat_bot_id = self.chat_task.bot_id
        message = db.ChatBotsHistory(
            chat_bot_id=chat_bot_id,
            type=message.type.name,
            text=message.text,
            user=message.user,
            destination=chat_id,
        )
        db.session.add(message)
        db.session.commit()

    def _get_chat_history(self, chat_id):
        chat_bot_id = self.chat_task.bot_id
        query = db.ChatBotsHistory.query\
            .filter(
                db.ChatBotsHistory.chat_bot_id == chat_bot_id,
                db.ChatBotsHistory.destination == chat_id
            )\
            .order_by(db.ChatBotsHistory.sent_at.desc())\
            .limit(self.MAX_DEPTH)

        result = [
            ChatBotMessage(
                rec.type,
                rec.text,
                rec.user,
                sent_at=rec.sent_at,
            )
            for rec in query
        ]
        result.reverse()
        return result


class ChatMemory:
    '''
    interface to work with individual chat
    '''
    def __init__(self, memory, chat_id):
        self.memory = memory
        self.chat_id = chat_id

        self.cached = False

    def get_history(self, cached=True):
        result = self.memory.get_chat_history(self.chat_id, cached=cached and self.cached)
        self.cached = True
        return result

    def add_to_history(self, message):
        self.memory.add_to_history(self.chat_id, message)

    def get_mode(self):
        return self.memory.get_mode(self.chat_id)

    def set_mode(self, mode):
        self.memory.set_mode(self.chat_id, mode)

    def hide_history(self, left_count):
        '''
        set date to start hiding messages
        '''
        self.memory.hide_history(self.chat_id, left_count)



import pandas as pd
from typing import List
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.handlers.query_utilities import SELECTQueryParser, SELECTQueryExecutor
from mindsdb.utilities import log
from mindsdb_sql.parser import ast

logger = log.getLogger(__name__)


class AQByUserLocationTable(APITable):
    """The Air Quality By User Location Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://aqicn.org/json-api/doc/#api-Geolocalized_Feed-GetHereFeed" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Air Quality Data

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'air_quality_user_location',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.aqicn_client.air_quality_user_location()

        self.check_res(res=response)

        df = pd.json_normalize(response["content"])

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["content"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "status",
            "data.aqi",
            "data.idx",
            "data.attributions",
            "data.city.geo",
            "data.city.name",
            "data.city.url",
            "data.city.location",
            "data.dominentpol",
            "data.iaqi.co.v",
            "data.iaqi.dew.v",
            "data.iaqi.h.v",
            "data.iaqi.no2.v",
            "data.iaqi.p.v",
            "data.iaqi.pm10.v",
            "data.iaqi.so2.v",
            "data.iaqi.t.v",
            "data.iaqi.w.v",
            "data.time.s",
            "data.time.tz",
            "data.time.v",
            "data.time.iso",
            "data.forecast.daily.o3",
            "data.forecast.daily.pm10",
            "data.forecast.daily.pm25",
            "data.debug.sync"
        ]


class AQByCityTable(APITable):
    """The Air Quality By City Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://aqicn.org/json-api/doc/#api-City_Feed-GetCityFeed" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Air Quality Data

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'air_quality_city',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'city':
                if op == '=':
                    search_params["city"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for city column.")
            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("city" in search_params)

        if not filter_flag:
            raise NotImplementedError("city column has to be present in where clause.")

        df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.aqicn_client.air_quality_city(search_params["city"])

        self.check_res(res=response)

        df = pd.json_normalize(response["content"])

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["content"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "status",
            "data.aqi",
            "data.idx",
            "data.attributions",
            "data.city.geo",
            "data.city.name",
            "data.city.url",
            "data.city.location",
            "data.dominentpol",
            "data.iaqi.co.v",
            "data.iaqi.dew.v",
            "data.iaqi.h.v",
            "data.iaqi.no2.v",
            "data.iaqi.o3.v",
            "data.iaqi.p.v",
            "data.iaqi.pm10.v",
            "data.iaqi.pm25.v",
            "data.iaqi.so2.v",
            "data.iaqi.t.v",
            "data.iaqi.w.v",
            "data.time.s",
            "data.time.tz",
            "data.time.v",
            "data.time.iso",
            "data.forecast.daily.o3",
            "data.forecast.daily.pm10",
            "data.forecast.daily.pm25",
            "data.debug.sync"
        ]


class AQByLatLngTable(APITable):
    """The Air Quality By Lat Lng Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://aqicn.org/json-api/doc/#api-Geolocalized_Feed-GetGeolocFeed" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Air Quality Data

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'air_quality_lat_lng',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'lat':
                if op == '=':
                    search_params["lat"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for lat column.")
            if arg1 == 'lng':
                if op == '=':
                    search_params["lng"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for lng column.")
            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("lat" in search_params) and ("lng" in search_params)

        if not filter_flag:
            raise NotImplementedError("lat and lng columns have to be present in where clause.")

        df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.aqicn_client.air_quality_lat_lng(search_params["lat"], search_params["lng"])

        self.check_res(res=response)

        df = pd.json_normalize(response["content"])

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["content"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "status",
            "data.aqi",
            "data.idx",
            "data.attributions",
            "data.city.geo",
            "data.city.name",
            "data.city.url",
            "data.city.location",
            "data.dominentpol",
            "data.iaqi.co.v",
            "data.iaqi.dew.v",
            "data.iaqi.h.v",
            "data.iaqi.no2.v",
            "data.iaqi.o3.v",
            "data.iaqi.p.v",
            "data.iaqi.pm10.v",
            "data.iaqi.pm25.v",
            "data.iaqi.so2.v",
            "data.iaqi.t.v",
            "data.iaqi.w.v",
            "data.time.s",
            "data.time.tz",
            "data.time.v",
            "data.time.iso",
            "data.forecast.daily.o3",
            "data.forecast.daily.pm10",
            "data.forecast.daily.pm25",
            "data.debug.sync"
        ]


class AQByNetworkStationTable(APITable):
    """The Air Quality By Network Station Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the https://aqicn.org/json-api/doc/#api-Search-SearchByName" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Air Quality Data

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'air_quality_station_by_name',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'name':
                if op == '=':
                    search_params["name"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for name column.")
            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("name" in search_params)

        if not filter_flag:
            raise NotImplementedError("name column have to be present in where clause.")

        df = pd.DataFrame(columns=self.get_columns())

        response = self.handler.aqicn_client.air_quality_station_by_name(search_params["name"])

        self.check_res(res=response)

        df = pd.json_normalize(response["content"])

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def check_res(self, res):
        if res["code"] != 200:
            raise Exception("Error fetching results - " + res["content"])

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "status",
            "data.aqi",
            "data.idx",
            "data.attributions",
            "data.city.geo",
            "data.city.name",
            "data.city.url",
            "data.city.location",
            "data.dominentpol",
            "data.iaqi.co.v",
            "data.iaqi.dew.v",
            "data.iaqi.h.v",
            "data.iaqi.no2.v",
            "data.iaqi.o3.v",
            "data.iaqi.p.v",
            "data.iaqi.pm10.v",
            "data.iaqi.pm25.v",
            "data.iaqi.so2.v",
            "data.iaqi.t.v",
            "data.iaqi.w.v",
            "data.time.s",
            "data.time.tz",
            "data.time.v",
            "data.time.iso",
            "data.forecast.daily.o3",
            "data.forecast.daily.pm10",
            "data.forecast.daily.pm25",
            "data.debug.sync"
        ]



from typing import Text, Dict

from mindsdb.utilities import log
from mindsdb_sql import parse_sql
from mindsdb.integrations.utilities.handlers.auth_utilities import MSGraphAPIAuthManager
from mindsdb.integrations.handlers.ms_teams_handler.settings import ms_teams_handler_config
from mindsdb.integrations.handlers.ms_teams_handler.ms_graph_api_teams_client import MSGraphAPITeamsClient

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.integrations.libs.api_handler import APIChatHandler
from mindsdb.integrations.handlers.ms_teams_handler.ms_teams_tables import ChannelsTable, ChannelMessagesTable, ChatsTable, ChatMessagesTable

logger = log.getLogger(__name__)


class MSTeamsHandler(APIChatHandler):
    """
    The Microsoft Teams handler implementation.
    """

    name = 'teams'

    def __init__(self, name: str, **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            **kwargs: arbitrary keyword arguments.
        """

        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.handler_storage = kwargs['handler_storage']
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        channels_data = ChannelsTable(self)
        self._register_table("channels", channels_data)

        channel_messages_data = ChannelMessagesTable(self)
        self._register_table("channel_messages", channel_messages_data)

        chats_data = ChatsTable(self)
        self._register_table("chats", chats_data)

        chat_messages_data = ChatMessagesTable(self)
        self._register_table("chat_messages", chat_messages_data)

    def connect(self) -> MSGraphAPITeamsClient:
        """
        Set up the connection required by the handler.

        Returns
        -------
        MSGraphAPITeamsClient
            Client object for accessing the Microsoft Graph API.
        """

        if self.is_connected and self.connection.check_connection():
            return self.connection

        # initialize the auth manager for the Microsoft Graph API
        ms_graph_api_auth_manager = MSGraphAPIAuthManager(
            handler_storage=self.handler_storage,
            scopes=self.connection_data.get('scopes', ms_teams_handler_config.DEFAULT_SCOPES),
            client_id=self.connection_data["client_id"],
            client_secret=self.connection_data["client_secret"],
            tenant_id=self.connection_data["tenant_id"],
            code=self.connection_data.get('code')
        )

        # get access token from the auth manager for the Microsoft Graph API
        access_token = ms_graph_api_auth_manager.get_access_token()

        # pass the access token to the client for access to the Microsoft Graph API
        self.connection = MSGraphAPITeamsClient(access_token)

        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.

        Returns
        -------
        StatusResponse
            Response object with the status of the connection.
        """

        response = StatusResponse(False)

        try:
            connection = self.connect()
            connection.check_connection()
            response.success = True
            response.copy_storage = True
        except Exception as e:
            logger.error(f'Error connecting to Microsoft Teams: {e}!')
            response.success = False
            response.error_message = str(e)

        self.is_connected = response.success

        return response

    def native_query(self, query: Text) -> StatusResponse:
        """
        Receive and process a raw query.

        Parameters
        ----------
        query: Text
            Query in the native format.

        Returns
        -------
        StatusResponse
            Response object with the result of the query.
        """

        ast = parse_sql(query, dialect="mindsdb")

        return self.query(ast)
    
    def get_chat_config(self) -> Dict:
        """
        Get the configuration for the chatbot.
        This method is required for the implementation of the chatbot.

        Returns
        -------
        Dict
            Configuration for the chatbot.
        """

        params = {
            'polling': {
                'type': 'message_count',
                'table': 'chats',
                'chat_id_col': 'id',
                'count_col': 'lastMessagePreview_id'
            },
            'chat_table': {
                'name': 'chat_messages',
                'chat_id_col': 'chatId',
                'username_col': 'from_user_displayName',
                'text_col': 'body_content',
                'time_col': 'createdDateTime',
            }
        }

        return params
    
    def get_my_user_name(self) -> Text:
        """
        Get the name of the signed in user.
        This method is required for the implementation of the chatbot.

        Returns
        -------
        Text
            Name of the signed in user.
        """        

        connection = self.connect()
        user_profile = connection.get_user_profile()
        
        return user_profile['displayName']


import unittest

from mindsdb.api.mysql.mysql_proxy.mysql_proxy import RESPONSE_TYPE
from mindsdb.integrations.handlers.trino_handler.trino_handler import TrinoHandler


class TrinoHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "host": "qa.analytics.quantum.site.gs.com",
                "port": "8090",
                "user": "dqsvcuat",
                "password": "",
                "catalog": "gsam_dev2imddata_elastic",
                "schema": "default",
                "service_name": "HTTP/qa.analytics.quantum.site.gs.com",
                "config_file_name": "test_trino_config.ini"
            }
        }
        cls.handler = TrinoHandler('test_trino_handler', **cls.kwargs)

    def test_0_canary(self):
        print('Running canary test')
        assert True
        print('Canary test ran successfully')

    def test_1_check_connection(self):
        conn_status = self.handler.check_connection()
        print('Trino connection status: ', conn_status)
        assert conn_status.get('success')

    def test_2_get_tables(self):
        tables = self.handler.get_tables()
        assert tables

    def test_3_describe_table(self):
        described = self.handler.get_columns("axioma_att_2021-12")
        assert described['type'] is not RESPONSE_TYPE.ERROR

    # TODO: complete tests implementation
    # def test_4_select_query(self):
    #     query = "SELECT * FROM data.test_mdb WHERE 'id'='1'"
    #     result = self.handler.query(query)
    #     assert result['type'] is RESPONSE_TYPE.TABLE
    #



import os
import datetime as dt
from typing import List
import pandas as pd
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from slack_sdk.socket_mode import SocketModeClient
from slack_sdk.socket_mode.request import SocketModeRequest
from slack_sdk.socket_mode.response import SocketModeResponse

from mindsdb.utilities import log
from mindsdb.utilities.config import Config

from mindsdb_sql.parser import ast
from mindsdb_sql.parser.ast import ASTNode
from mindsdb.integrations.libs.api_handler import APIChatHandler, APIResource, FuncParser
from mindsdb.integrations.utilities.sql_utils import FilterCondition, FilterOperator

from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

logger = log.getLogger(__name__)

DATE_FORMAT = '%Y-%m-%d %H:%M:%S'


class SlackChannelListsTable(APIResource):

    def list(self, **kwargs) -> pd.DataFrame:

        client = self.handler.connect()

        channels = client.conversations_list(types="public_channel,private_channel")['channels']

        for channel in channels:
            channel['created_at'] = dt.datetime.fromtimestamp(channel['created'])
            channel['updated_at'] = dt.datetime.fromtimestamp(channel['updated'] / 1000)

        return pd.DataFrame(channels, columns=self.get_columns())

    def get_columns(self) -> List[str]:
        return [
            'id',
            'name',
            'created_at',
            'updated_at'
        ]


class SlackChannelsTable(APIResource):

    def list(self,
             conditions: List[FilterCondition] = None,
             limit: int = None,
             **kwargs) -> pd.DataFrame:
        """
        Retrieves the data from the channel using SlackAPI

        Returns:
            conversation_history, pd.DataFrame
        """

        client = self.handler.connect()

        # override the default function
        def parse_utc_date(date_str):
            date_obj = dt.datetime.fromisoformat(date_str).replace(tzinfo=dt.timezone.utc)
            return date_obj

        # Get the channels list and ids
        channels = client.conversations_list(types="public_channel,private_channel")['channels']
        channel_ids = {c['name']: c['id'] for c in channels}
        
        # Extract comparison conditions from the query
        channel_name = None

        # Build the filters and parameters for the query
        params = {}

        for condition in conditions:
            value = condition.value
            op = condition.op

            if condition.column == 'channel':
                if value in channel_ids:
                    params['channel'] = channel_ids[value]
                    channel_name = value
                    condition.applied = True
                else:
                    raise ValueError(f"Channel '{value}' not found")

            # Is this used?
            # elif condition.column == 'limit':
            #     if op == FilterOperator.EQUAL:
            #         params['limit'] = int(value)
            #         condition.applied = True
            #     else:
            #         raise NotImplementedError(f'Unknown op: {op}')

            elif condition.column == 'created_at' and value is not None:
                date = parse_utc_date(value)
                if op == FilterOperator.GREATER_THAN:
                    params['oldest'] = date.timestamp() + 1
                elif op == FilterOperator.GREATER_THAN_OR_EQUAL:
                    params['oldest'] = date.timestamp()
                elif op == FilterOperator.LESS_THAN_OR_EQUAL:
                    params['latest'] = date.timestamp()
                else:
                    continue
                condition.applied = True

        if limit:
            params['limit'] = limit

        # Retrieve the conversation history
        result = client.conversations_history(**params)

        # convert SlackResponse object to pandas DataFrame
        result = pd.DataFrame(result['messages'], columns=self.get_columns())

        # Remove null rows from the result
        result = result[result['text'].notnull()]

        # Add the selected channel to the dataframe
        result['channel'] = channel_name

        # translate the time stamp into a 'created_at' field
        result['created_at'] = pd.to_datetime(result['ts'].astype(float), unit='s').dt.strftime('%Y-%m-%d %H:%M:%S')

        return result

    def get_columns(self) -> List[str]:
        return [
            'channel',
            'client_msg_id',
            'type',
            'subtype',
            'ts',
            'created_at',
            'user',
            'text',
            'attachments',
            'files',
            'reactions',
            'thread_ts',
            'reply_count',
            'reply_users_count',
            'latest_reply',
            'reply_users'
        ]

    def insert(self, query):
        """
        Inserts the message in the Slack Channel

        Args:
            channel_name
            message
        """
        client = self.handler.connect()
    
        # get column names and values from the query
        columns = [col.name for col in query.columns]
        for row in query.values:
            params = dict(zip(columns, row))

            # check if required parameters are provided
            if 'channel' not in params or 'text' not in params:
                raise Exception("To insert data into Slack, you need to provide the 'channel' and 'text' parameters.")

            # post message to Slack channel
            try:
                response = client.chat_postMessage(
                    channel=params['channel'],
                    text=params['text']
                )
            except SlackApiError as e:
                raise Exception(f"Error posting message to Slack channel '{params['channel']}': {e.response['error']}")
            
            inserted_id = response['ts']
            params['ts'] = inserted_id

    def update(self, query: ast.Update):
        """
        Updates the message in the Slack Channel

        Args:
            updated message
            channel_name
            ts  [TimeStamp -> Can be found by running select command, the entire result will be printed in the terminal]
        """
        client = self.handler.connect()

        # Get the channels list and ids
        channels = client.conversations_list(types="public_channel,private_channel")['channels']
        channel_ids = {c['name']: c['id'] for c in channels}

        # Extract comparison conditions from the query
        conditions = extract_comparison_conditions(query.where)

        filters = []
        params = {}

        keys = list(query.update_columns.keys())

        # Build the filters and parameters for the query
        for op, arg1, arg2 in conditions:
            if arg1 == 'channel':
                if arg2 in channel_ids:
                    params['channel'] = channel_ids[arg2]
                else:
                    raise ValueError(f"Channel '{arg2}' not found")

            if keys[0] == 'text':
                params['text'] =  str(query.update_columns['text'])
            else:
                raise ValueError(f"Message '{arg2}' not found")

            if arg1 == 'ts':
                if op == '=':
                    params['ts'] = float(arg2)
                else:
                    raise NotImplementedError(f'Unknown op: {op}')
            else:
                filters.append([op, arg1, arg2])

        # check if required parameters are provided
        if 'channel' not in params or 'ts' not in params or 'text' not in params:
            raise Exception("To update a message in Slack, you need to provide the 'channel', 'ts', and 'text' parameters.")

        # update message in Slack channel
        try:
            response = client.chat_update(
                channel=params['channel'],
                ts=str(params['ts']),
                text=params['text'].strip()
            )
        except SlackApiError as e:
            raise Exception(f"Error updating message in Slack channel '{params['channel']}' with timestamp '{params['ts']}' and message '{params['text']}': {e.response['error']}")

    def delete(self, query: ASTNode):
        """
        Deletes the message in the Slack Channel

        Args:
            channel_name
            ts  [TimeStamp -> Can be found by running select command, the entire result will be printed in the terminal]
        """
        client = self.handler.connect()

        # Get the channels list and ids
        channels = client.conversations_list(types="public_channel,private_channel")['channels']
        channel_ids = {c['name']: c['id'] for c in channels}
        # Extract comparison conditions from the query
        conditions = extract_comparison_conditions(query.where)

        filters = []
        params = {}

        # Build the filters and parameters for the query
        for op, arg1, arg2 in conditions:
            if arg1 == 'channel':
                if arg2 in channel_ids:
                    params['channel'] = channel_ids[arg2]
                else:
                    raise ValueError(f"Channel '{arg2}' not found")

            if arg1 == 'ts':
                if op == '=':
                    params['ts'] = float(arg2)
                else:
                    raise NotImplementedError(f'Unknown op: {op}')

            else:
                filters.append([op, arg1, arg2])

        # check if required parameters are provided
        if 'channel' not in params or 'ts' not in params:
            raise Exception("To delete a message from Slack, you need to provide the 'channel' and 'ts' parameters.")

        # delete message from Slack channel
        try:
            response = client.chat_delete(
                channel=params['channel'],
                ts=params['ts']
            )
            
        except SlackApiError as e:
            raise Exception(f"Error deleting message from Slack channel '{params['channel']}' with timestamp '{params['ts']}': {e.response['error']}")


class SlackHandler(APIChatHandler):
    """
    A class for handling connections and interactions with Slack API.
    Agrs:
        bot_token(str): The bot token for the Slack app.
    """

    def __init__(self, name=None, **kwargs):
        """
        Initializes the connection by checking all the params are provided by the user.
        """
        super().__init__(name)

        args = kwargs.get('connection_data', {})
        self.connection_args = {}
        handler_config = Config().get('slack_handler', {})
        for k in ['token', 'app_token']:
            if k in args:
                self.connection_args[k] = args[k]
            elif f'SLACK_{k.upper()}' in os.environ:
                self.connection_args[k] = os.environ[f'SLACK_{k.upper()}']
            elif k in handler_config:
                self.connection_args[k] = handler_config[k]
        self.api = None
        self.is_connected = False
        
        channels = SlackChannelsTable(self)
        self._register_table('channels', channels)

        channel_lists = SlackChannelListsTable(self)
        self._register_table('channel_lists', channel_lists)

        self._socket_mode_client = None

    def get_chat_config(self):
        params = {
            'polling': {
                'type': 'realtime',
                'table_name': 'channels'
            },
            'chat_table': {
                'name': 'channels',
                'chat_id_col': 'channel',
                'username_col': 'user',
                'text_col': 'text',
                'time_col': 'thread_ts',
            }
        }
        return params

    def get_my_user_name(self):
        # TODO
        api = self.connect()
        resp = api.users_profile_get()
        return resp.data['profile']['bot_id']

    def subscribe(self, stop_event, callback, table_name, **kwargs):
        if table_name != 'channels':
            raise RuntimeError(f'Table not supported: {table_name}')

        self._socket_mode_client = SocketModeClient(
            # This app-level token will be used only for establishing a connection
            app_token=self.connection_args['app_token'],  # xapp-A111-222-xyz
            # You will be using this WebClient for performing Web API calls in listeners
            web_client=WebClient(token=self.connection_args['token']),  # xoxb-111-222-xyz
        )

        def _process_websocket_message(client: SocketModeClient, request: SocketModeRequest):
            # Acknowledge the request
            response = SocketModeResponse(envelope_id=request.envelope_id)
            client.send_socket_mode_response(response)

            if request.type != 'events_api':
                return

            payload_event = request.payload['event']
            if payload_event['type'] != 'message':
                return
            if 'subtype' in payload_event:
                # Don't respond to message_changed, message_deleted, etc.
                return
            if payload_event['channel_type'] != 'im':
                # Only support IMs currently.
                return
            if 'bot_id' in payload_event:
                # A bot sent this message.
                return

            key = {
                'channel': payload_event['channel'],
            }
            row = {
                'text': payload_event['text'],
                'user': payload_event['user'],
            }

            callback(row, key)

        self._socket_mode_client.socket_mode_request_listeners.append(_process_websocket_message)
        self._socket_mode_client.connect()

        stop_event.wait()

        self._socket_mode_client.close()


    def create_connection(self):
        """
        Creates a WebClient object to connect to the Slack API token stored in the connection_args attribute.
        """

        client = WebClient(token=self.connection_args['token'])
        return client
    
    def connect(self):
        """
        Authenticate with the Slack API using the token stored in the `token` attribute.
        """
        if self.is_connected is True:
            return self.api

        self.api = self.create_connection()
        return self.api

    def check_connection(self):
        """
        Checks the connection by calling auth_test()
        """
        response = StatusResponse(False)

        try:
            api = self.connect()

            # Call API method to check the connection
            api.auth_test()

            # check app_token
            if 'app_token' in self.connection_args:
                socket_client = SocketModeClient(
                    app_token=self.connection_args['app_token'],
                    web_client=api
                )
                socket_client.connect()
                socket_client.disconnect()
            
            response.success = True
        except SlackApiError as e:
            response.error_message = f'Error connecting to Slack Api: {e.response["error"]}. Check token.'
            logger.error(response.error_message)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query_string: str = None):
        """
        Parses the query with FuncParser and calls call_slack_api and returns the result of the query as a Response object.
        """
        method_name, params = FuncParser().from_string(query_string)

        df = self.call_slack_api(method_name, params)

        return Response(
            RESPONSE_TYPE.TABLE,
            data_frame=df
        )

    def call_slack_api(self, method_name: str = None, params: dict = None):
        """
        Calls specific method specified.

        Args:
            method_name: to call specific method
            params: parameters to call the method

        Returns:
            List of dictionaries as a result of the method call
        """
        api = self.connect()
        method = getattr(api, method_name)

        try:
            result = method(**params)

        except SlackApiError as e:
            error = f"Error calling method '{method_name}' with params '{params}': {e.response['error']}"
            logger.error(error)
            raise e

        if 'channels' in result:
            result['channels'] = self.convert_channel_data(result['channels'])

        return [result]

    def convert_channel_data(self, channels: List[dict]):
        """
        Convert the list of channel dictionaries to a format that can be easily used in the data pipeline.

        Args:
            channels: A list of channel dictionaries.

        Returns:
            A list of channel dictionaries with modified keys and values.
        """

        new_channels = []
        for channel in channels:
            new_channel = {
                'id': channel['id'],
                'name': channel['name'],
                'created': dt.datetime.fromtimestamp(float(channel['created']))
            }
            new_channels.append(new_channel)
        return new_channels



from typing import Iterable, List, Optional

import pandas as pd
from mindsdb_sql import parse_sql, Identifier
from mindsdb_sql.planner.utils import query_traversal


from mindsdb.utilities import log

logger = log.getLogger(__name__)


class SQLAgent:

    def __init__(
            self,
            command_executor,
            database: str,
            include_tables: Optional[List[str]] = None,
            ignore_tables: Optional[List[str]] = None,
            sample_rows_in_table_info: int = 3,
    ):
        self._database = database
        self._command_executor = command_executor
        self._integration_controller = command_executor.session.integration_controller

        self._sample_rows_in_table_info = int(sample_rows_in_table_info)

        self._tables_to_include = include_tables
        self._tables_to_ignore = []
        self._database = database
        if not self._tables_to_include:
            # ignore_tables and include_tables should not be used together.
            # include_tables takes priority if it's set.
            self._tables_to_ignore = ignore_tables or []

    def _call_engine(self, query: str, database=None):
        # switch database

        ast_query = parse_sql(query.strip('`'))
        self._check_tables(ast_query)

        if database is None:
            database = self._database

        ret = self._command_executor.execute_command(
            ast_query,
            database_name=database
        )
        return ret

    def _check_tables(self, ast_query):

        def _check_f(node, is_table=None, **kwargs):
            if is_table and isinstance(node, Identifier):
                table = node.parts[-1]
                if table not in self._tables_to_include:
                    ValueError(f"Table {table} not found. Available tables: {', '.join(self._tables_to_include)}")

        query_traversal(ast_query, _check_f)

    def get_usable_table_names(self) -> Iterable[str]:
        if self._tables_to_include:
            return self._tables_to_include

        ret = self._call_engine('show databases;')
        dbs = [lst[0] for lst in ret.data if lst[0] != 'information_schema']
        usable_tables = []
        for db in dbs:
            if db != 'mindsdb' and db == self._database:
                try:
                    ret = self._call_engine('show tables', database=db)
                    tables = [lst[0] for lst in ret.data if lst[0] != 'information_schema']
                    for table in tables:
                        # By default, include all tables in a database unless expilcitly ignored.
                        table_name = f'{db}.{table}'
                        if table_name not in self._tables_to_ignore:
                            usable_tables.append(table_name)
                except Exception as e:
                    logger.warning('Unable to get tables for %s: %s', db, str(e))

        return usable_tables

    def get_table_info(self, table_names: Optional[List[str]] = None) -> str:
        """ Get information about specified tables.
        Follows best practices as specified in: Rajkumar et al, 2022 (https://arxiv.org/abs/2204.00498)
        If `sample_rows_in_table_info`, the specified number of sample rows will be
        appended to each table description. This can increase performance as demonstrated in the paper.
        """
        all_table_names = self.get_usable_table_names()
        if table_names is not None:
            missing_tables = set(table_names).difference(all_table_names)
            if missing_tables:
                raise ValueError(f"table_names {missing_tables} not found in database")
            all_table_names = table_names

        tables = []
        for table in all_table_names:
            table_info = self._get_single_table_info(table)
            tables.append(table_info)

        final_str = "\n\n".join(tables)
        return final_str

    def get_table_columns(self, table_name: str) -> List[str]:
        controller = self._integration_controller
        cols_df = controller.get_data_handler(self._database).get_columns(table_name).data_frame
        return cols_df['Field'].to_list()

    def _get_single_table_info(self, table_str: str) -> str:
        controller = self._integration_controller
        integration, table_name = table_str.split('.')
        cols_df = controller.get_data_handler(integration).get_columns(table_name).data_frame
        fields = cols_df['Field'].to_list()
        dtypes = cols_df['Type'].to_list()

        info = f'Table named `{table_name}`\n'
        info += f"\n/* Sample with first {self._sample_rows_in_table_info} rows from table `{table_str}`:\n"
        info += "\t".join([field for field in fields])
        info += self._get_sample_rows(table_str, fields) + "\n*/"
        info += '\nColumn data types: ' + ",\t".join(
            [f'`{field}` : `{dtype}`' for field, dtype in zip(fields, dtypes)]) + '\n'  # noqa
        return info

    def _get_sample_rows(self, table: str, fields: List[str]) -> str:
        command = f"select {','.join(fields)} from {table} limit {self._sample_rows_in_table_info};"
        try:
            ret = self._call_engine(command)
            sample_rows = ret.data
            sample_rows = list(
                map(lambda ls: [str(i) if len(str(i)) < 100 else str[:100] + '...' for i in ls], sample_rows))
            sample_rows_str = "\n" + "\n".join(["\t".join(row) for row in sample_rows])
        except Exception:
            sample_rows_str = "\n" + "\t [error] Couldn't retrieve sample rows!"

        return sample_rows_str

    def query(self, command: str, fetch: str = "all") -> str:
        """Execute a SQL command and return a string representing the results.
        If the statement returns rows, a string of the results is returned.
        If the statement returns no rows, an empty string is returned.
        """

        def _tidy(result: List) -> str:
            return '\n'.join(['\t'.join([str(value) for value in row]) for row in result])

        def _repr_result(ret):
            limit_rows = 30

            columns_str = ', '.join([repr(col.name) for col in ret.columns])
            res = f'Output columns: {columns_str}\n'

            if len(ret.data) > limit_rows:
                df = pd.DataFrame(ret.data, columns=[col.name for col in ret.columns])

                res += f'Result has {len(ret.data)} rows. Description of data:\n'
                res += str(df.describe(include='all')) + '\n\n'
                res += f'First {limit_rows} rows:\n'

            else:
                res += 'Result:\n'

            res += _tidy(ret.data[:limit_rows])
            return res

        ret = self._call_engine(command)
        if fetch == "all":
            result = _repr_result(ret)
        elif fetch == "one":
            result = _tidy(ret.data[0])
        else:
            raise ValueError("Fetch parameter must be either 'one' or 'all'")
        return str(result)

    def get_table_info_safe(self, table_names: Optional[List[str]] = None) -> str:
        try:
            return self.get_table_info(table_names)
        except Exception as e:
            return f"Error: {e}"

    def query_safe(self, command: str, fetch: str = "all") -> str:
        try:
            return self.query(command, fetch)
        except Exception as e:
            return f"Error: {e}"



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Dremio server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The port that Dremio is running on.'
    },
    username={
        'type': ARG_TYPE.STR,
        'description': 'The username used to authenticate with the Dremio server.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the Dremio server.',
        'secret': True
    }
)

connection_args_example = OrderedDict(
    host='localhost',
    database=9047,
    username='admin',
    password='password'
)



from typing import Optional

import pandas as pd
import oracledb
from oracledb import connect, Connection, makedsn

from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)

oracledb.defaults.fetch_lobs = False  # return LOBs directly as strings or bytes

logger = log.getLogger(__name__)

class OracleHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Microsoft SQL Server statements.
    """

    name = "oracle"

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        super().__init__(name)
        self.host = connection_data.get("host")
        self.port = int(connection_data.get("port") or 1521)
        self.sid = connection_data.get("sid")
        self.service_name = connection_data.get("service_name")
        self.user = connection_data.get("user")
        self.password = connection_data.get("password")
        self.disable_oob = bool(connection_data.get("disable_oob"))

        self.auth_mode = None
        if 'auth_mode' in connection_data:
            mode_name = 'AUTH_MODE_' + connection_data['auth_mode'].upper()
            if not hasattr(oracledb, mode_name):
                raise ValueError(f'Unknown auth mode: {mode_name}')
            self.auth_mode = getattr(oracledb, mode_name)

        if self.sid is None and self.service_name is None:
            raise ValueError("Either 'sid' or 'service_name' must be given")
        if self.sid and self.service_name:
            raise ValueError("Only one of 'sid' or 'service_name' must be given")

        if self.sid:
            self.dsn = makedsn(host=self.host, port=self.port, sid=self.sid)
        else:
            self.dsn = makedsn(host=self.host, port=self.port, service_name=self.service_name)

        self.connection = None
        self.is_connected = False

    def connect(self) -> Connection:
        if self.is_connected is True:
            return self.connection
        d = None  # default suitable for Linux OS
        oracledb.init_oracle_client(lib_dir=d)
        connection = connect(
            user=self.user, password=self.password, dsn=self.dsn,
            disable_oob=self.disable_oob, mode=self.auth_mode,
        )

        self.is_connected = True
        self.connection = connection
        return self.connection

    def disconnect(self):
        if self.is_connected:
            self.connection.close()
        self.is_connected = False
        return

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the database
        :return: success status and error message if error occurs
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            con = self.connect()
            con.ping()
            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to Oracle DB {self.dsn}, {e}!")
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False
        return response

    def native_query(self, query: str) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run
        :return: returns the records from the current recordset
        """
        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cur:
            try:
                cur.execute(query)
                result = cur.fetchall()
                if result:
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame(
                            result,
                            columns=[row[0] for row in cur.description],
                        ),
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)

                connection.commit()
            except Exception as e:
                logger.error(f"Error running query: {query} on {self.dsn}!")
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e),
                )
                connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement.
        """
        renderer = SqlalchemyRender("oracle")
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        List all tables in Oracle DB owned by the current user.
        """
        query = """
            SELECT table_name
            FROM user_tables
            ORDER BY 1
        """
        return self.native_query(query)

    def get_columns(self, table_name: str) -> Response:
        """
        Show details about the table.
        """
        query = f"""
            SELECT 
                column_name,
                data_type
            FROM USER_TAB_COLUMNS
            WHERE table_name = '{table_name}'
        """
        result = self.native_query(query)
        return result



import unittest

from mindsdb.integrations.handlers.surrealdb_handler.surrealdb_handler import SurrealDBHandler
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE


class SurrealdbHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "host": "localhost",
                "port": "8000",
                "user": "admin",
                "password": "password",
                "namespace": "test",
                "database": "test"
            }
        }
        cls.handler = SurrealDBHandler('test_surrealdb_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_create_table(self):
        res = self.handler.native_query("CREATE person SET name = 'Tobie', company = 'SurrealDB', "
                                        "skills = ['Rust', 'Go', 'JavaScript'];")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_2_insert(self):
        res = self.handler.native_query("UPDATE person SET name = 'Jamie'")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_3_select_query(self):
        query = "SELECT * FROM person"
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_4_get_columns(self):
        columns = self.handler.get_columns('person')
        assert columns.type is not RESPONSE_TYPE.ERROR

    def test_5_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_6_drop_table(self):
        res = self.handler.native_query("REMOVE table person")
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_7_disconnect(self):
        assert self.handler.disconnect() is None


if __name__ == '__main__':
    unittest.main()



from langchain.agents import AgentType

SUPPORTED_PROVIDERS = {'openai', 'anthropic', 'anyscale', 'litellm', 'ollama'}
# Chat models
ANTHROPIC_CHAT_MODELS = {
    'claude-3-opus-20240229',
    'claude-3-sonnet-20240229',
    'claude-3-haiku-20240307',
    'claude-2.1',
    'claude-2.0',
    'claude-instant-1.2'
}
# See https://ollama.com/library.
OLLAMA_CHAT_MODELS = {
    "gemma",
    "llama2",
    "mistral",
    "mixtral",
    "llava",
    "neural-chat",
    "codellama",
    "dolphin-mixtral",
    "qwen",
    "llama2-uncensored",
    "mistral-openorca",
    "deepseek-coder",
    "nous-hermes2",
    "phi",
    "orca-mini",
    "dolphin-mistral",
    "wizard-vicuna-uncensored",
    "vicuna",
    "tinydolphin",
    "llama2-chinese",
    "openhermes",
    "zephyr",
    "nomic-embed-text",
    "tinyllama",
    "openchat",
    "wizardcoder",
    "phind-codellama",
    "starcoder",
    "yi",
    "orca2",
    "falcon",
    "starcoder2",
    "wizard-math",
    "dolphin-phi",
    "nous-hermes",
    "starling-lm",
    "stable-code",
    "medllama2",
    "bakllava",
    "codeup",
    "wizardlm-uncensored",
    "solar",
    "everythinglm",
    "sqlcoder",
    "nous-hermes2-mixtral",
    "stable-beluga",
    "yarn-mistral",
    "samantha-mistral",
    "stablelm2",
    "meditron",
    "stablelm-zephyr",
    "magicoder",
    "yarn-llama2",
    "wizard-vicuna",
    "llama-pro",
    "deepseek-llm",
    "codebooga",
    "mistrallite",
    "dolphincoder",
    "nexusraven",
    "open-orca-platypus2",
    "all-minilm",
    "goliath",
    "notux",
    "alfred",
    "megadolphin",
    "xwinlm",
    "wizardlm",
    "duckdb-nsql",
    "notus"
}
DEFAULT_ASSISTANT_COLUMN = 'answer'
DEFAULT_AGENT_TIMEOUT_SECONDS = 300
# These should require no additional arguments.
DEFAULT_AGENT_TOOLS = []  
DEFAULT_AGENT_TYPE = AgentType.CONVERSATIONAL_REACT_DESCRIPTION
DEFAULT_MAX_ITERATIONS = 10
DEFAULT_MAX_TOKENS = 2048
DEFAULT_MODEL_NAME = 'gpt-4-0125-preview'
DEFAULT_USER_COLUMN = 'question'
DEFAULT_EMBEDDINGS_MODEL_PROVIDER = 'openai'


import pandas as pd
from typing import Text, List, Dict, Any

from mindsdb.utilities import log
from mindsdb_sql.parser import ast
from mindsdb.integrations.libs.api_handler import APITable

from mindsdb.integrations.handlers.ms_teams_handler.settings import ms_teams_handler_config

from mindsdb.integrations.utilities.handlers.query_utilities.insert_query_utilities import INSERTQueryParser
from mindsdb.integrations.utilities.handlers.query_utilities.select_query_utilities import SELECTQueryParser, SELECTQueryExecutor

logger = log.getLogger(__name__)


class ChatsTable(APITable):
    """
    The Microsoft Teams Chats Table implementation.
    """
    
    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls data from the "GET /chats" and "GET /chats/{chat_id} Microsoft Graph API endpoints.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query.

        Returns
        -------
        pd.DataFrame
            Microsoft Teams Chats matching the query.

        Raises
        ------
        ValueError
            If the query contains an unsupported target (column).

        NotImplementedError
            If the query contains an unsupported condition.
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'chats',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        # only the = operator is supported for the id column
        id = None
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'id':
                if op == "=":
                    id = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for id column.")
        
        chats_df = pd.json_normalize(self.get_chats(id), sep='_')

        # if the id is given, remove the id column from the where conditions as it has already been evaluated in the API call (get_chats)
        if id:
            where_conditions = [where_condition for where_condition in where_conditions if where_condition[1] not in ['id']]

        select_statement_executor = SELECTQueryExecutor(
            chats_df,
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit if query.limit else None
        )

        chats_df = select_statement_executor.execute_query()

        return chats_df
    
    def get_chats(self, chat_id: Text = None) -> List[Dict[Text, Any]]:
        """
        Calls the API client to get the chats from the Microsoft Graph API.

        Parameters
        ----------
        chat_id: Text
            The chat id to get the chat from.

        Returns
        -------
        List[Dict[Text, Any]]
            The chats from the Microsoft Graph API.
        """

        api_client = self.handler.connect()

        # if the chat_id is given, get the chat with that id from the API
        if chat_id:
            chats = [api_client.get_chat(chat_id)]
        # if the chat_id is not given, get all the chats
        else:
            chats = api_client.get_chats()

        for chat in chats:
            last_message_preview = chat.get("lastMessagePreview")

            # keep only the lastMessagePreview_id and lastMessagePreview_createdDateTime columns
            if last_message_preview:
                chat["lastMessagePreview_id"] = last_message_preview.get("id")
                chat["lastMessagePreview_createdDateTime"] = last_message_preview.get("createdDateTime")
                del chat["lastMessagePreview"]
                del chat["lastMessagePreview@odata.context"]

        return chats

    def get_columns(self) -> List[Text]:
        """
        Returns the columns of the Chats Table.

        Returns
        -------
        List[Text]
            The columns of the Chats Table.
        """

        return ms_teams_handler_config.CHATS_TABLE_COLUMNS
    
class ChatMessagesTable(APITable):
    """
    The Microsoft Teams Chat Messages Table implementation.
    """
    
    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls data from the "GET /chats/{chat_id}/messages" and "GET /chats/{chat_id}/messages/{message_id}" Microsoft Graph API endpoints.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query.

        Returns
        -------
        pd.DataFrame
            Microsoft Teams Chat Messages matching the query.

        Raises
        ------
        ValueError
            If the query contains an unsupported target (column).

        NotImplementedError
            If the query contains an unsupported condition.
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'chat_messages',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        # only the = operator is supported for the id and chatId columns
        chat_id, message_id = None, None
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'id':
                if op == "=":
                    message_id = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for id column.")
                
            if arg1 == 'chatId':
                if op == "=":
                    chat_id = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for chatId column.")
        
        messages_df = pd.json_normalize(self.get_messages(chat_id, message_id), sep='_')
    
        # if both chat_id and message_id are given, remove the id and chatId columns from the where conditions as they have already been evaluated in the API call (get_messages)
        if chat_id and message_id:
            where_conditions = [where_condition for where_condition in where_conditions if where_condition[1] not in ['id', 'chatId']]
        # if only the chat_id is given, remove the chatId column from the where conditions as it has already been evaluated in the API call (get_messages)
        elif chat_id:
            where_conditions = [where_condition for where_condition in where_conditions if where_condition[1] not in ['chatId']]

        select_statement_executor = SELECTQueryExecutor(
            messages_df,
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit if query.limit else None
        )

        messages_df = select_statement_executor.execute_query()

        return messages_df
    
    def get_messages(self, chat_id: Text = None, message_id: Text = None) -> List[Dict[Text, Any]]:
        """
        Calls the API client to get the messages from the Microsoft Graph API.
        If all parameters are None, it will return all the messages from all the chats.
        If only the chat_id is given, it will return all the messages from that chat.

        Parameters
        ----------
        chat_id: Text
            The chat id to get the messages from.

        message_id: Text
            The message id to get the message from.

        Returns
        -------
        List[Dict[Text, Any]]
            The messages from the Microsoft Graph API.
        """

        api_client = self.handler.connect()

        # if both chat_id and message_id are given, get the message with that id from the API
        if message_id and chat_id:
            chat_messages = [api_client.get_chat_message(chat_id, message_id)]
        # if only the chat_id is given, get all the messages from that chat
        elif chat_id:
            chat_messages = api_client.get_chat_messages(chat_id)
        # if no parameters are given or only the message_id is given, get all the messages from all the chats
        else:
            chat_messages = api_client.get_all_chat_messages()

        return chat_messages

    def insert(self, query: ast.Insert) -> None:
        """
        Inserts data into the "POST /chats/{chat_id}/messages" Microsoft Graph API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query.

        Returns
        -------
        None

        Raises
        ------
        UnsupportedColumnException
            If the query contains an unsupported column.

        MandatoryColumnException
            If the query is missing a mandatory column.

        ColumnCountMismatchException
            If the number of columns does not match the number of values.
        """

        # only the chatId, subject and body_content columns are supported
        # chatId and body_content are mandatory
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=["chatId", "subject", "body_content"],
            mandatory_columns=["chatId", "body_content"],
        )

        messages_to_send = insert_statement_parser.parse_query()

        self.send_messages(messages_to_send)

    def send_messages(self, messages_to_send: List[Dict[Text, Any]]) -> None:
        """
        Calls the API client to send the messages to the Microsoft Graph API.

        Parameters
        ----------
        messages_to_send: List[Dict[Text, Any]]
            The messages to send to the Microsoft Graph API.

        Returns
        -------
        None
        """

        api_client = self.handler.connect()

        # send each message through the API to the chat with the given id
        for message in messages_to_send:
            api_client.send_chat_message(
                chat_id=message["chatId"],
                message=message["body_content"],
                subject=message.get("subject")
            )

    def get_columns(self) -> List[Text]:
        """
        Returns the columns of the Chat Messages Table.

        Returns
        -------
        List[Text]
            The columns of the Chat Messages Table.
        """

        return ms_teams_handler_config.CHAT_MESSAGES_TABLE_COLUMNS

    class ChatMessageRepliesTable(APITable):
        """
        The Microsoft Chat Message Replies Table implementation.
        """
        pass

class ChannelsTable(APITable):
    """
    The Microsoft Channels Table implementation.
    """

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls data from the "GET /teams/{group_id}/channels" and "GET teams/{group_id}/channels/{channel_id}" Microsoft Graph API endpoints.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query.

        Returns
        -------
        pd.DataFrame
            Microsoft Teams Channels matching the query.

        Raises
        ------
        ValueError
            If the query contains an unsupported target (column).

        NotImplementedError
            If the query contains an unsupported condition.
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'channels',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        # only the = operator is supported for the id and teamId columns
        channel_id, team_id = None, None
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'id':
                if op == "=":
                    channel_id = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for id column.")
                
            if arg1 == 'teamId':
                if op == "=":
                    team_id = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for teamId column.")
                
        channels_df = pd.json_normalize(self.get_channels(channel_id, team_id))

        # if both channel_id and team_id are given, remove the id and teamId columns from the where conditions as they have already been evaluated in the API call (get_channels)
        if channel_id and team_id:
            where_conditions = [where_condition for where_condition in where_conditions if where_condition[1] not in ['id', 'teamId']]

        select_statement_executor = SELECTQueryExecutor(
            channels_df,
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit if query.limit else None
        )

        channels_df = select_statement_executor.execute_query()

        return channels_df

    def get_channels(self, channel_id: Text = None, team_id: Text = None) -> List[Dict[Text, Any]]:
        """
        Calls the API client to get the channels from the Microsoft Graph API.
        If all parameters are None, it will return all the channels from all the teams.

        Parameters
        ----------
        channel_id: Text
            The channel id to get the channel from.

        team_id: Text
            The team id to get the channels from.

        Returns
        -------
        List[Dict[Text, Any]]
            The channels from the Microsoft Graph API.
        """

        api_client = self.handler.connect()

        # if both channel_id and team_id are given, get the channel with that id from the API
        if channel_id and team_id:
            return [api_client.get_channel(team_id, channel_id)]
        # if no parameter are given or only the team_id is given, get all the channels
        else:
            return api_client.get_channels()
    
    def get_columns(self) -> List[Text]:
        """
        Returns the columns of the Channels Table.

        Returns
        -------
        List[Text]
            The columns of the Channels Table.
        """

        return ms_teams_handler_config.CHANNELS_TABLE_COLUMNS

class ChannelMessagesTable(APITable):
    """
    The Microsoft Teams Channel Messages Table implementation.       
    """
     
    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls data from the "GET /teams/{group_id}/channels/{channel_id}/messages" and "GET /teams/{group_id}/channels/{channel_id}/messages/{message_id}" Microsoft Graph API endpoints.

        Parameters
        ----------
        query: ast.Select
            Given SQL SELECT query.

        Returns
        -------
        pd.DataFrame
            Microsoft Teams Channel Messages matching the query.

        Raises
        ------
        ValueError
            If the query contains an unsupported target (column).

        NotImplementedError
            If the query contains an unsupported condition.
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'channel_messages',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        # only the = operator is supported for the id, channelIdentity_teamId and channelIdentity_channelId columns
        team_id, channel_id, message_id = None, None, None
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'id':
                if op == "=":
                    message_id = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for id column.")

            if arg1 == 'channelIdentity_teamId':
                if op == "=":
                    team_id = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for id column.")
                
            if arg1 == 'channelIdentity_channelId':
                if op == "=":
                    channel_id = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for teamId column.")
                
        messages_df = pd.json_normalize(self.get_messages(team_id, channel_id, message_id), sep='_')

        # if all parameters are given, remove the id, channelIdentity_teamId and channelIdentity_channelId columns from the where conditions as they have already been evaluated in the API call (get_messages)
        if team_id and channel_id and message_id:
            where_conditions = [where_condition for where_condition in where_conditions if where_condition[1] not in ['id', 'channelIdentity_teamId', 'channelIdentity_channelId']]

        select_statement_executor = SELECTQueryExecutor(
            messages_df,
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit if query.limit else None
        )

        messages_df = select_statement_executor.execute_query()

        return messages_df
    
    def get_messages(self, team_id: Text = None, channel_id: Text = None, message_id: Text = None) -> List[Dict[Text, Any]]:
        """
        Calls the API client to get the messages from the Microsoft Graph API.
        If all parameters are None, it will return all the messages from all the channels from all the teams.

        Parameters
        ----------
        team_id: Text
            The team id to get the messages from.

        channel_id: Text
            The channel id to get the messages from.

        message_id: Text
            The message id to get the message from.
        """

        api_client = self.handler.connect()

        # if all parameters are given, get the message with that id from that channel from that team from the API
        if message_id and channel_id and team_id:
            channel_message = api_client.get_channel_message(team_id, channel_id, message_id)
            # add the missing eventDetail attribute to the channel message
            channel_message['eventDetail'] = {
                '@odata.type': None, 
                'visibleHistoryStartDateTime': None, 
                'members': None,
                'channelId': None,
                'channelDisplayName': None,
                'initiator': {
                    'application': {
                        '@odata.type': None, 
                        'id': None,
                        'displayName': None,
                        'applicationIdentityType': None
                    },
                    'device': None, 
                    'user': {
                        '@odata.type': None, 
                        'id': None,
                        'displayName': None,
                        'userIdentityType': None,
                        'tenantId': None
                    }
                }
            }

            return [channel_message]
        # for any other combination of parameters, get all the messages from all the channels from all the teams
        else:
            return api_client.get_channel_messages()
    
    def insert(self, query: ast.Insert) -> None:
        """
        Inserts data into the "POST /teams/{group_id}/channels/{channel_id}/messages" Microsoft Graph API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query.

        Returns
        -------
        None

        Raises
        ------
        UnsupportedColumnException
            If the query contains an unsupported column.

        MandatoryColumnException
            If the query is missing a mandatory column.

        ColumnCountMismatchException
            If the number of columns does not match the number of values.
        """

        # only the channelIdentity_teamId, channelIdentity_channelId, subject and body_content columns are supported
        # channelIdentity_teamId and channelIdentity_channelId are mandatory
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=["channelIdentity_teamId", "channelIdentity_channelId", "subject", "body_content"],
            mandatory_columns=["channelIdentity_teamId", "channelIdentity_channelId", "body_content"],
        )

        messages_to_send = insert_statement_parser.parse_query()

        self.send_messages(messages_to_send)

    def send_messages(self, messages_to_send: List[Dict[Text, Any]]) -> None:
        """
        Calls the API client to send the messages to the Microsoft Graph API.
        
        Parameters
        ----------
        messages_to_send: List[Dict[Text, Any]]
            The messages to send to the Microsoft Graph API.

        Returns
        -------
        None

        Raises
        ------
        None
        """

        api_client = self.handler.connect()

        # send each message through the API to the channel with the given id from the team with the given id
        for message in messages_to_send:
            api_client.send_channel_message(
                group_id=message["channelIdentity_teamId"],
                channel_id=message["channelIdentity_channelId"],
                message=message["body_content"],
                subject=message.get("subject")
            )
    
    def get_columns(self) -> List[Text]:
        """
        Returns the columns of the Channel Messages Table.

        Returns
        -------
        list
            The columns of the Channel Messages Table.
        """

        return ms_teams_handler_config.CHANNEL_MESSAGES_TABLE_COLUMNS

class ChannelMessageRepliesTable(APITable):
    """
    The Microsoft Teams Channel Message Replies Table implementation.
    """
    pass



"""
    Wrapper around MindsDB's executor and integration controller following the implementation of the original
    langchain.sql_database.SQLDatabase class to partly replicate its behavior.
"""
from typing import Any, Iterable, List, Optional

from langchain.sql_database import SQLDatabase

from mindsdb.interfaces.skills.skill_tool import skill_tool
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class MindsDBSQL(SQLDatabase):
    """ Can't modify signature, as LangChain does a Pydantic check."""
    def __init__(
        self,
        engine=None,
        database: Optional[str] = 'mindsdb',
        metadata: Optional[Any] = None,
        ignore_tables: Optional[List[str]] = None,
        include_tables: Optional[List[str]] = None,
        sample_rows_in_table_info: int = 3,
        schema: Optional[str] = None,
        indexes_in_table_info: bool = False,
        custom_table_info: Optional[dict] = None,
        view_support: Optional[bool] = True,
    ):
        # Some args above are not used in this class, but are kept for compatibility

        self._sql_agent = skill_tool.get_sql_agent(
            database,
            include_tables,
            ignore_tables,
            sample_rows_in_table_info
        )

    @property
    def dialect(self) -> str:
        return 'mindsdb'

    def get_usable_table_names(self) -> Iterable[str]:
        return self._sql_agent.get_usable_table_names()

    def get_table_info_no_throw(self, table_names: Optional[List[str]] = None) -> str:
        return self._sql_agent.get_table_info_safe(table_names)

    def run_no_throw(self, command: str, fetch: str = "all") -> str:
        return self._sql_agent.query_safe(command)


import mindsdb.interfaces.storage.db as db
from mindsdb.api.postgres.postgres_proxy.postgres_proxy import PostgresProxyHandler
from mindsdb.utilities import log


def start(verbose=False):
    logger = log.getLogger(__name__)
    logger.info("Postgres API is starting..")
    db.init()

    PostgresProxyHandler.startProxy()



import re
from typing import Optional

import duckdb
import aerospike
import pandas as pd
# from sqlalchemy import create_engine

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode

# from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)


class AerospikeHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Solr SQL statements.
    """
    name = 'aerospike'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'aerospike'
        self.connection_data = connection_data
        self.kwargs = kwargs
        if not self.connection_data.get('host'):
            raise Exception("The host parameter should be provided!")
        if not self.connection_data.get('port'):
            raise Exception("The port parameter should be provided!")

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """
        if self.is_connected is True:
            return self.connection

        user = self.connection_data.get('user', None)
        password = self.connection_data.get('password', None)
        config = {
            'user': user,
            'password': password,
            'hosts': [(self.connection_data.get('host'), self.connection_data.get('port'))],
        }
        connection = aerospike.client(config).connect()
        self.is_connected = True
        self.connection = connection.connect()
        return self.connection

    def disconnect(self):
        """
        Close any existing connections.
        """
        if self.is_connected is False:
            return
        self.connection.close()
        self.is_connected = False
        return

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the Aerospike database
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            response.error_message = str(e)

        if response.success is True and need_to_close:
            self.disconnect()
        if response.success is False and self.is_connected is True:
            self.is_connected = False
        return response

    def native_query(self, query: str) -> Response:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        connection = self.connect()

        try:
            # where is not supported
            selected_bins, aero_ns, aero_set = self.parse_aql_query(query)
            aero_ns = aero_ns.lower()
            aero_set = aero_set.lower()
            scan = connection.scan(aero_ns.lower(), aero_set.lower())
            res = scan.results()
            data_df = pd.DataFrame.from_records([r[2] for r in res])
            if ' where ' in query or ' WHERE ' in query or '*' not in selected_bins:
                new_query = re.sub(r'FROM [\w\.]+', 'FROM ' + 'data_df', query, 1)
                new_query = new_query.replace(f'{aero_set}.', '')
                data_df = duckdb.query(new_query).to_df()

            response = Response(
                RESPONSE_TYPE.TABLE,
                data_df
            )
        except Exception as e:
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement.
        """
        return self.native_query(query.to_string())

    def parse_aql_query(self, aql_query):
        # Split the AQL query into tokens
        tokens = [t.replace(',', '').upper() for t in re.split(r'\s+', aql_query)]
        # Extract the relevant components
        select_index = tokens.index("SELECT")
        from_index = tokens.index("FROM")
        # where_index = tokens.index("WHERE")

        selected_bins = tokens[select_index + 1:from_index]
        namespace_set = tokens[from_index + 1]
        aero_ns, aero_set = namespace_set.split('.') if '.' in namespace_set else None, namespace_set
        if not aero_ns:
            aero_ns = self.connection_data.get('namespace')
        # filter_condition = " ".join(tokens[where_index + 1:])
        return selected_bins, aero_ns, aero_set

    def get_tables(self) -> Response:
        """
        Get a list with all of the tables in Aerospike
        """
        need_to_close = self.is_connected is False
        connection = self.connect()

        data_lst = []
        request = "sets"

        try:
            for node, (err, res) in list(connection.info_all(request).items()):
                if res:
                    entries = [entry.strip() for entry in res.strip().split(';') if entry.strip()]
                    for entry in entries:
                        data = [d for d in entry.split('=') if ':set' in d or ':objects' in d]
                        ele = [None, None, None]
                        for d in data:
                            if ':set' in d:
                                ele[0] = d.split(':')[0]
                            if ':objects' in d:
                                ele[1] = d.split(':')[0]
                            if d[0] or d[1]:
                                ele[2] = request
                        data_lst.append(ele)

            response = Response(
                RESPONSE_TYPE.TABLE,
                pd.DataFrame(data_lst, columns=['table_schema', 'table_name', 'table_type'])
            )
        except Exception as e:
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        if need_to_close is True:
            self.disconnect()

        return response

    def get_columns(self, table_name: str) -> Response:
        """
        Show details about the table
        """
        need_to_close = self.is_connected is False
        connection = self.connect()

        column_df = pd.DataFrame([], columns=['column_name', 'data_type'])

        try:
            response_table = self.get_tables()
            df = response_table.data_frame
            if not len(df):
                return column_df
            df = df[df['table_name'] == table_name]
            tbl_dtl_arr = df.iloc[0][['table_schema', 'table_name']]
            scan = connection.scan(tbl_dtl_arr[0], tbl_dtl_arr[1])
            res = scan.results()
            data_df = pd.DataFrame.from_records([r[2] for r in res])
            column_df = pd.DataFrame(data_df.dtypes).reset_index()
            column_df.columns = ['column_name', 'data_type']
            response = Response(
                RESPONSE_TYPE.TABLE,
                column_df
            )
        except Exception as e:
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        if need_to_close is True:
            self.disconnect()

        return response
